from supervisely.app.singleton import Singleton


class JinjaWidgets(metaclass=Singleton):
    """Singleton context provider for Jinja2 templates with widget scripts and auto-assigned widget IDs."""

    def __init__(self, auto_widget_id=True):
        """
        :param auto_widget_id: If True, auto-generate widget IDs when None.
        :type auto_widget_id: bool
        """
        self.auto_widget_id = auto_widget_id
        self.context = {}
        self.context["__no_html_mode__"] = auto_widget_id
        self.context["__widget_scripts__"] = {}



# coding: utf-8

from __future__ import annotations

import uuid
from typing import Dict, List, Optional

from bidict import bidict

from supervisely._utils import take_with_default
from supervisely.annotation.label import LabelJsonFields
from supervisely.annotation.obj_class import ObjClass
from supervisely.collection.key_indexed_collection import KeyObject
from supervisely.geometry.constants import (
    CLASS_ID,
    CREATED_AT,
    LABELER_LOGIN,
    UPDATED_AT,
)
from supervisely.project.project_meta import ProjectMeta
from supervisely.project.project_settings import LabelingInterface
from supervisely.video_annotation.constants import ID, KEY, OBJECTS_MAP
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_tag import VideoTag
from supervisely.video_annotation.video_tag_collection import VideoTagCollection


class VideoObject(KeyObject):
    """Tracked object across video frames; obj_class plus optional tags. Immutable."""

    def __init__(
        self,
        obj_class: ObjClass,
        tags: Optional[VideoTagCollection] = None,
        key: Optional[uuid.UUID] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[str] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
    ):
        """
        :param obj_class: Object class (e.g. 'car' with Rectangle geometry).
        :type obj_class: :class:`~supervisely.annotation.obj_class.ObjClass`
        :param tags: Tags for this object.
        :type tags: :class:`~supervisely.video_annotation.video_tag_collection.VideoTagCollection`, optional
        :param key: UUID key. Auto-generated if not provided.
        :type key: uuid.UUID, optional
        :param class_id: Server-side class ID.
        :type class_id: int, optional
        :param labeler_login: Login of user who created the object.
        :type labeler_login: str, optional
        :param updated_at: Last modification timestamp.
        :type updated_at: str, optional
        :param created_at: Creation timestamp.
        :type created_at: str, optional

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                obj_class_car = sly.ObjClass('car', sly.Rectangle)
                video_obj_car = sly.VideoObject(obj_class_car)
                video_obj_car_json = video_obj_car.to_json()
                print(video_obj_car_json)
        """
        self.labeler_login = labeler_login
        self.updated_at = updated_at
        self.created_at = created_at

        self._obj_class = obj_class
        self._key = take_with_default(key, uuid.uuid4())
        self._tags = take_with_default(tags, VideoTagCollection())
        self._class_id = take_with_default(class_id, None)

    def _add_creation_info(self, d):
        if self.labeler_login is not None:
            d[LABELER_LOGIN] = self.labeler_login
        if self.updated_at is not None:
            d[UPDATED_AT] = self.updated_at
        if self.created_at is not None:
            d[CREATED_AT] = self.created_at

    @property
    def obj_class(self) -> ObjClass:
        """
        ObjClass of the current VideoObject.

        :returns: ObjClass object
        :rtype: :class:`~supervisely.annotation.obj_class.ObjClass`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                obj_class_car = sly.ObjClass('car', sly.Rectangle)
                video_obj_car = sly.VideoObject(obj_class_car)
                obj_car_json = video_obj_car.obj_class.to_json()
                print(obj_car_json)
                # Output: {
                #     "title": "car",
                #     "shape": "rectangle",
                #     "color": "#8A0F7B",
                #     "geometry_config": {},
                #     "hotkey": ""
                # }
        """
        return self._obj_class

    def key(self) -> uuid.UUID:
        """
        Object key.

        :returns: Object key
        :rtype: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :rtype: uuid.UUID

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                obj_class_car = sly.ObjClass('car', sly.Rectangle)
                video_obj_car = sly.VideoObject(obj_class_car)
                key = video_obj_car.key()
                print(key)
                # Output: 158e6cf4f4ac4c639fc6994aad127c16
        """

        return self._key

    @property
    def tags(self) -> VideoTagCollection:
        """
        VideoTagCollection of the current VideoObject.

        :returns: VideoTagCollection object
        :rtype: :class:`~supervisely.video_annotation.video_tag_collection.VideoTagCollection`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                # Create VideoTagCollection
                meta_car = sly.TagMeta('car_tag', sly.TagValueType.ANY_STRING)
                from supervisely.video_annotation.video_tag import VideoTag
                car_tag = VideoTag(meta_car, value='acura')
                from supervisely.video_annotation.video_tag_collection import VideoTagCollection
                video_tags = VideoTagCollection([car_tag])

                # Create VideoObject
                obj_class_car = sly.ObjClass('car', sly.Rectangle)
                video_obj_car = sly.VideoObject(obj_class_car, video_tags)

                # Get VideoTagCollection
                tags = video_obj_car.tags
                tags_json = tags.to_json()
                print(tags_json)
                # Output: [
                #     {
                #         "name": "car_tag",
                #         "value": "acura",
                #         "key": "4f82fbcab74c44259d7a0e29d604602e"
                #     }
                # ]
        """
        return self._tags.clone()

    @property
    def class_id(self) -> int:
        """
        Object class ID.

        :returns: Object class ID.
        :rtype: int

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                obj_class_car = sly.ObjClass('car', sly.Rectangle)
                video_obj_car = sly.VideoObject(obj_class_car)
                class_id = video_obj_car.class_id
        """

        return self._class_id

    def add_tag(self, tag: VideoTag) -> VideoObject:
        """
        Adds VideoTag to the current VideoObject.

        :param tag: VideoTag to be added.
        :type tag: :class:`~supervisely.video_annotation.video_tag.VideoTag`
        :returns: VideoObject object
        :rtype: :class:`~supervisely.video_annotation.video_object.VideoObject`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                # Create VideoObject
                obj_class_car = sly.ObjClass('car', sly.Rectangle)
                video_obj_car = sly.VideoObject(obj_class_car)

                # Create VideoTag
                meta_car = sly.TagMeta('car_tag', sly.TagValueType.ANY_STRING)
                from supervisely.video_annotation.video_tag import VideoTag
                car_tag = VideoTag(meta_car, value='acura')

                # Add VideoTag
                new_obj_car = video_obj_car.add_tag(car_tag)
                new_obj_car_json = new_obj_car.to_json()
                print(new_obj_car_json)
                # Output: {
                #     "key": "1ab52285ee634c93b724fa655b785eae",
                #     "classTitle": "car",
                #     "tags": [
                #         {
                #             "name": "car_tag",
                #             "value": "acura",
                #             "key": "d9e52b275e074c538f162a6d679aed9e"
                #         }
                #     ]
                # }
        """
        return self.clone(tags=self._tags.add(tag))

    def add_tags(self, tags: List[VideoTag]) -> VideoObject:
        """
        Adds VideoTags to the current VideoObject.

        :param tag: List of VideoTags to be added.
        :type tag: List[:class:`~supervisely.video_annotation.video_tag.VideoTag`]
        :returns: VideoObject object
        :rtype: :class:`~supervisely.video_annotation.video_object.VideoObject`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                # Create VideoObject
                obj_class_vehicle = sly.ObjClass('vehicle', sly.Rectangle)
                video_obj_vehicle = sly.VideoObject(obj_class_vehicle)

                # Create VideoTags
                from supervisely.video_annotation.video_tag import VideoTag
                meta_car = sly.TagMeta('car_tag', sly.TagValueType.ANY_STRING)
                car_tag = VideoTag(meta_car, value='acura')
                meta_bus = sly.TagMeta('bus_tag', sly.TagValueType.ANY_STRING)
                bus_tag = VideoTag(meta_bus, value='volvo')

                # Add VideoTags
                new_obj_vehicle = video_obj_vehicle.add_tags([car_tag, bus_tag])
                new_obj_vehicle_json = new_obj_vehicle.to_json()
                print(new_obj_vehicle_json)
                # Output: {
                #     "key": "94055c5e8cb146368f627fc608fb6b44",
                #     "classTitle": "vehicle",
                #     "tags": [
                #         {
                #             "name": "car_tag",
                #             "value": "acura",
                #             "key": "6679f47c96734565919fbffc278532a1"
                #         },
                #         {
                #             "name": "bus_tag",
                #             "value": "volvo",
                #             "key": "d0a60dc929de491a85a09fea59adb818"
                #         }
                #     ]
                # }
        """
        return self.clone(tags=self._tags.add_items(tags))

    def to_json(self, key_id_map: Optional[KeyIdMap] = None) -> Dict:
        """
        Convert the VideoObject to a json dict. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :returns: Json format as a dict
        :rtype: dict

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                obj_class_car = sly.ObjClass('vehicle', sly.Rectangle)
                video_obj_car = sly.VideoObject(obj_class_car)

                obj_car_json = video_obj_car.to_json()
                print(obj_car_json)
                # Output: {
                #     "key": "ce26e77a45bc45e88e3e17da1672d01f",
                #     "classTitle": "vehicle",
                #     "tags": []
                # }
        """
        data_json = {
            KEY: self.key().hex,
            LabelJsonFields.OBJ_CLASS_NAME: self.obj_class.name,
            LabelJsonFields.TAGS: self.tags.to_json(key_id_map),
        }

        if key_id_map is not None:
            item_id = key_id_map.get_object_id(self.key())
            if item_id is not None:
                data_json[ID] = item_id

        self._add_creation_info(data_json)
        return data_json

    @classmethod
    def from_json(
        cls, data: Dict, project_meta: ProjectMeta, key_id_map: Optional[KeyIdMap] = None
    ) -> VideoObject:
        """
        Convert a json dict to VideoObject. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: Dict in json format.
        :type data: dict
        :param project_meta: Input ProjectMeta object.
        :type project_meta: :class:`~supervisely.project.project_meta.ProjectMeta`
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :raises RuntimeError: if object class name is not found in the given project meta
        :returns: VideoObject object
        :rtype: :class:`~supervisely.video_annotation.video_object.VideoObject`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                obj_car_json = {"classTitle": "vehicle", "tags": []}
                obj_class_car = sly.ObjClass('vehicle', sly.Rectangle)
                obj_collection = sly.ObjClassCollection([obj_class_car])
                meta = sly.ProjectMeta(obj_classes=obj_collection)

                video_obj_from_json = sly.VideoObject.from_json(obj_car_json, meta)
        """
        obj_class_name = data[LabelJsonFields.OBJ_CLASS_NAME]
        obj_class = project_meta.get_obj_class(obj_class_name)
        if obj_class is None:
            raise RuntimeError(
                f"Failed to deserialize a object from JSON: class name {obj_class_name!r} "
                f"was not found in the given project meta."
            )

        is_multiview = False
        try:
            if project_meta.labeling_interface == LabelingInterface.MULTIVIEW:
                is_multiview = True
        except AttributeError:
            is_multiview = False

        raw_id = data.get(ID, None)

        if not is_multiview:
            key = uuid.UUID(data[KEY]) if KEY in data else uuid.uuid4()
            if key_id_map is not None:
                key_id_map.add_object(key, raw_id)
        else:
            if KEY in data:
                key = uuid.UUID(data[KEY])
            elif key_id_map is not None and isinstance(raw_id, int):
                existing_key = key_id_map.get_object_key(raw_id)
                if isinstance(existing_key, uuid.UUID):
                    key = existing_key
                else:
                    key = uuid.uuid4()
            else:
                key = uuid.uuid4()

            if key_id_map is not None and isinstance(raw_id, int):
                if key_id_map.get_object_id(key) is None:
                    key_id_map.add_object(key, raw_id)

        class_id = data.get(CLASS_ID, None)
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)

        return cls(
            obj_class=obj_class,
            key=key,
            tags=VideoTagCollection.from_json(data[LabelJsonFields.TAGS], project_meta.tag_metas),
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

    def clone(
        self,
        obj_class: Optional[ObjClass] = None,
        tags: Optional[VideoTagCollection] = None,
        key: Optional[uuid.UUID] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[str] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
    ) -> VideoObject:
        """
        Makes a copy of VideoObject with new fields, if fields are given, otherwise it will use fields of the original VideoObject.

        :param obj_class: ObjClass object.
        :type obj_class: :class:`~supervisely.annotation.obj_class.ObjClass`, optional
        :param tags: VideoTagCollection object.
        :type tags: :class:`~supervisely.video_annotation.video_tag_collection.VideoTagCollection`, optional
        :param key: UUID key associated with the object.
        :type key: uuid.UUID, optional
        :param class_id: ID of ObjClass to which VideoObject belongs.
        :type class_id: int, optional
        :param labeler_login: Login of the user who created VideoObject.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when VideoObject was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when VideoObject was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional
        :returns: VideoObject object
        :rtype: :class:`~supervisely.video_annotation.video_object.VideoObject`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.video_annotation.video_tag import VideoTag
                from supervisely.video_annotation.video_tag_collection import VideoTagCollection

                obj_class_car = sly.ObjClass('vehicle', sly.Rectangle)
                video_obj_car = sly.VideoObject(obj_class_car)

                # New ObjClass to clone
                obj_class_bus = sly.ObjClass('bus', sly.Rectangle)

                # New VideoTagCollection to clone
                meta_car = sly.TagMeta('car_tag', sly.TagValueType.ANY_STRING)
                car_tag = VideoTag(meta_car, value='acura')
                tags = VideoTagCollection([car_tag])
                # Clone
                new_obj_vehicle = video_obj_car.clone(obj_class=obj_class_bus, tags=tags)
                new_obj_vehicle_json = new_obj_vehicle.to_json()
                print(new_obj_vehicle_json)
                # Output: {
                #     "key": "39ae5b9ce1ca405c9f53544374b3f5be",
                #     "classTitle": "bus",
                #     "tags": [
                #         {
                #             "name": "car_tag",
                #             "value": "acura",
                #             "key": "3119b6e38fe24fe7a220e881154fd9ba"
                #         }
                #     ]
                # }
        """
        return self.__class__(
            obj_class=take_with_default(obj_class, self.obj_class),
            key=take_with_default(key, self._key),
            tags=take_with_default(tags, self.tags),
            class_id=take_with_default(class_id, self.class_id),
            labeler_login=take_with_default(labeler_login, self.labeler_login),
            updated_at=take_with_default(updated_at, self.updated_at),
            created_at=take_with_default(created_at, self.created_at),
        )



# coding: utf-8

# docs
from __future__ import annotations

import base64
import gzip
import tempfile
from collections import OrderedDict
from typing import Dict, List, Literal, Optional, Tuple, Union

import numpy as np

from supervisely.sly_logger import logger
from supervisely._utils import unwrap_if_numpy
from supervisely.geometry.constants import (
    CLASS_ID,
    CREATED_AT,
    DATA,
    GEOMETRY_SHAPE,
    GEOMETRY_TYPE,
    ID,
    LABELER_LOGIN,
    MASK_3D,
    SPACE,
    SPACE_DIRECTIONS,
    SPACE_ORIGIN,
    UPDATED_AT,
)
from supervisely.geometry.geometry import Geometry
from supervisely.io.fs import get_file_ext, get_file_name, remove_dir
from supervisely.io.json import JsonSerializable

if not hasattr(np, "bool"):
    np.bool = np.bool_


class PointVolume(JsonSerializable):
    """3D point coordinate (x, y, z) in volume space. Immutable."""

    def __init__(self, x: Union[int, float], y: Union[int, float], z: Union[int, float]):
        """
        PointVolume (x, y, z) determines position of Mask3D. It locates the first sample.
        :class:`~supervisely.geometry.mask_3d.PointVolume` object is immutable.

        :param x: Position of :class:`~supervisely.geometry.mask_3d.PointVolume` object on X-axis.
        :type x: int or float
        :param y: Position of :class:`~supervisely.geometry.mask_3d.PointVolume` object on Y-axis.
        :type y: int or float
        :param z: Position of :class:`~supervisely.geometry.mask_3d.PointVolume` object on Z-axis.
        :type z: int or float

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                x = 100
                y = 200
                z = 2
                loc = sly.PointVolume(x, y, z)
        """
        self._x = round(unwrap_if_numpy(x))
        self._y = round(unwrap_if_numpy(y))
        self._z = round(unwrap_if_numpy(z))

    @property
    def x(self) -> int:
        """
        Position of PointVolume on X-axis.

        :returns: X of :class:`~supervisely.geometry.mask_3d.PointVolume`
        :rtype: int

        :Usage Example:

            .. code-block:: python

                print(loc.x)
                # Output: 100
        """
        return self._x

    @property
    def y(self) -> int:
        """
        Position of PointVolume on Y-axis.

        :returns: Y of :class:`~supervisely.geometry.mask_3d.PointVolume`
        :rtype: int

        :Usage Example:

            .. code-block:: python

                print(loc.y)
                # Output: 200
        """
        return self._y

    @property
    def z(self) -> int:
        """
        Position of PointVolume on Z-axis.

        :returns: Z of :class:`~supervisely.geometry.mask_3d.PointVolume`
        :rtype: int

        :Usage Example:

            .. code-block:: python

                print(loc.z)
                # Output: 2
        """
        return self._z

    def to_json(self) -> Dict:
        """
        Convert the PointVolume to a json dict.

        :returns: Json format as a dict
        :rtype: dict

        :Usage Example:

            .. code-block:: python

                loc_json = loc.to_json()
                print(loc_json)
                # Output: {
                #           "space_origin": [
                #                            200,
                #                            200,
                #                            100
                #                           ]
                #         }

        """

        packed_obj = {SPACE_ORIGIN: [self.x, self.y, self.z]}
        return packed_obj

    @classmethod
    def from_json(cls, packed_obj) -> PointVolume:
        """
        Convert a json dict to PointVolume.

        :param data: PointVolume in json format as a dict.
        :type data: dict
        :returns: :class:`~supervisely.geometry.mask_3d.PointVolume` object
        :rtype: :class:`~supervisely.geometry.mask_3d.PointVolume`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                loc_json = {"space_origin": [200, 200, 100]}

                loc = sly.PointVolume.from_json(loc_json)
        """
        return cls(
            x=packed_obj["space_origin"][0],
            y=packed_obj["space_origin"][1],
            z=packed_obj["space_origin"][2],
        )


class Mask3D(Geometry):
    """3D volumetric mask (voxel data). Immutable."""

    def __init__(
        self,
        data: np.ndarray,
        sly_id: Optional[int] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[str] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
        volume_header: Optional[Dict] = None,
        convert_to_ras: bool = True,
    ):
        """
        Mask 3D geometry for a single :class:`~supervisely.annotation.label.Label`. :class:`~supervisely.geometry.mask_3d.Mask3D` object is immutable.

        :param data: Mask 3D mask data. Must be a numpy array with only 2 unique values: [0, 1] or [0, 255] or [False, True].
        :type data: np.ndarray
        :param sly_id: Mask 3D ID in Supervisely server.
        :type sly_id: int, optional
        :param class_id: ID of ObjClass to which Mask 3D belongs.
        :type class_id: int, optional
        :param labeler_login: Login of the user who created Mask 3D.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when Mask 3D was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when Mask 3D was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional
        :param volume_header: NRRD header dictionary. Optional.
        :type volume_header: dict, optional
        :param convert_to_ras: If True, converts the mask to RAS orientation. Default is True.
        :type convert_to_ras: bool, optional
        :raises ValueError: if data is not bool or no pixels set to True in data

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                # Create simple Mask 3D
                mask3d = np.zeros((3, 3, 3), dtype=np.bool_)
                mask3d[0:2, 0:2, 0:2] = True

                shape = sly.Mask3D(mask3d)

                print(shape.data)
                # Output:
                #    [[[ True  True False]
                #      [ True  True False]
                #      [False False False]]

                #     [[ True  True False]
                #      [ True  True False]
                #      [False False False]]

                #     [[False False False]
                #      [False False False]
                #      [False False False]]]
        """
        super().__init__(
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

        if not isinstance(data, np.ndarray):
            raise TypeError('Mask 3D "data" argument must be numpy array object!')

        data_dims = len(data.shape)
        if data_dims != 3:
            raise ValueError(
                f'Mask 3D "data" argument must be a 3-dimensional numpy array. Instead got {data_dims} dimensions'
            )

        if data.dtype != np.bool:
            if list(np.unique(data)) not in [[0, 1], [0, 255]]:
                raise ValueError(
                    f"Mask 3D mask data values must be one of: [0 1], [0 255], [False True]. Instead got {np.unique(data)}."
                )

            if list(np.unique(data)) == [0, 1]:
                data = np.array(data, dtype=bool)
            elif list(np.unique(data)) == [0, 255]:
                data = np.array(data / 255, dtype=bool)

        self.data = data
        self._space_origin = None
        self._space = None
        self._space_directions = None

        if volume_header is not None:
            self.set_volume_space_meta(volume_header)
            if self.space is not None and self.space != "right-anterior-superior":
                if convert_to_ras:
                    self.orient_ras()
                else:
                    logger.debug(
                        "Mask3D is not in RAS orientation. It is recommended to use RAS orientation for 3D masks."
                    )

    @property
    def space_origin(self) -> Optional[List[float]]:
        """
        Get the space origin of the Mask3D as a list of floats.

        :returns: Space origin of the :class:`~supervisely.geometry.mask_3d.Mask3D`.
        :rtype: List[float] or None
        """
        if self._space_origin is not None:
            return [self._space_origin.x, self._space_origin.y, self._space_origin.z]
        return None

    @space_origin.setter
    def space_origin(self, value: Union[PointVolume, List[float], np.array]):
        """
        Set the space origin of the Mask3D.

        :param value: Space origin of the :class:`~supervisely.geometry.mask_3d.Mask3D`. If provided as a list or array, it should contain 3 floats in the order [x, y, z].
        :type value: :class:`~supervisely.geometry.mask_3d.PointVolume` or List[float]
        """
        if isinstance(value, PointVolume):
            self._space_origin = value
        elif isinstance(value, list) and len(value) == 3:
            self._space_origin = PointVolume(x=value[0], y=value[1], z=value[2])
        elif isinstance(value, np.ndarray) and value.shape == (3,):
            self._space_origin = PointVolume(x=value[0], y=value[1], z=value[2])
        else:
            raise ValueError("Space origin must be a PointVolume or a list of 3 floats.")

    @property
    def space(self) -> Optional[str]:
        """
        Get the space of the Mask3D.

        :returns: Space of the :class:`~supervisely.geometry.mask_3d.Mask3D`.
        :rtype: str
        """
        return self._space

    @space.setter
    def space(self, value: str):
        """
        Set the space of the Mask3D.

        :param value: Space of the :class:`~supervisely.geometry.mask_3d.Mask3D`.
        :type value: str
        """
        if not isinstance(value, str):
            raise ValueError("Space must be a string.")
        self._space = value

    @property
    def space_directions(self) -> Optional[List[List[float]]]:
        """
        Get the space directions of the Mask3D.

        :returns: Space directions of the :class:`~supervisely.geometry.mask_3d.Mask3D`.
        :rtype: List[List[float]]
        """
        return self._space_directions

    @space_directions.setter
    def space_directions(self, value: Union[List[List[float]], np.ndarray]):
        """
        Set the space directions of the Mask3D.

        :param value: Space directions of the :class:`~supervisely.geometry.mask_3d.Mask3D`. Should be a 3x3 array-like structure.
        :type value: List[List[float]] or np.ndarray
        """
        if isinstance(value, np.ndarray):
            if value.shape != (3, 3):
                raise ValueError("Space directions must be a 3x3 array.")
            self._space_directions = value.tolist()
        elif (
            isinstance(value, list)
            and len(value) == 3
            and all(isinstance(row, (list, np.ndarray)) and len(row) == 3 for row in value)
        ):
            self._space_directions = [list(row) for row in value]
        else:
            raise ValueError("Space directions must be a 3x3 array or list of lists.")

    @staticmethod
    def geometry_name():
        """Return geometry name"""
        return "mask_3d"

    @staticmethod
    def from_file(figure, file_path: str):
        """
        Load figure geometry from file.

        :param figure: Spatial figure
        :type figure: :class:`~supervisely.geometry.volume_figure.VolumeFigure`
        :param file_path: Path to nrrd file with data
        :type file_path: str
        """
        mask3d = Mask3D.create_from_file(file_path)
        figure._set_3d_geometry(mask3d)
        path_without_filename = "/".join(file_path.split("/")[:-1])
        remove_dir(path_without_filename)

    @classmethod
    def create_from_file(cls, file_path: str) -> Mask3D:
        """
        Creates Mask3D geometry from file.

        :param file_path: Path to nrrd file with data
        :type file_path: str
        """
        from supervisely.volume.volume import read_nrrd_serie_volume_np

        mask3d_data, meta = read_nrrd_serie_volume_np(file_path)
        direction = np.array(meta["directions"]).reshape(3, 3)
        spacing = np.array(meta["spacing"])
        space_directions = (direction.T * spacing[:, None]).tolist()
        mask3d_header = {
            "space": "right-anterior-superior",
            "space directions": space_directions,
            "space origin": meta.get("origin", None),
        }

        geometry = cls(data=mask3d_data, volume_header=mask3d_header)

        fields_to_check = ["space", "space_directions", "space_origin"]
        if any([getattr(geometry, value) is None for value in fields_to_check]):
            header_keys = ["'space'", "'space directions'", "'space origin'"]
            logger.debug(
                f"The Mask3D geometry created from the file '{file_path}' doesn't contain optional space attributes that have similar names to {', '.join(header_keys)}. To set the values for these attributes, you can use information from the Volume associated with this figure object."
            )
        return geometry

    @classmethod
    def from_bytes(cls, geometry_bytes: bytes) -> Mask3D:
        """
        Create a Mask3D geometry object from bytes.

        :param geometry_bytes: NRRD file represented as bytes.
        :type geometry_bytes: bytes
        :returns: A Mask3D geometry object.
        :rtype: :class:`~supervisely.geometry.mask_3d.Mask3D`
        """
        with tempfile.NamedTemporaryFile(delete=True, suffix=".nrrd") as temp_file:
            temp_file.write(geometry_bytes)
            return cls.create_from_file(temp_file.name)

    def to_json(self) -> Dict:
        """
        Convert the Mask 3D to a json dict.

        :returns: Json format as a dict
        :rtype: Dict

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                mask = np.array([[[1 1 0]
                                [1 1 0]
                                [0 0 0]]
                                [[1 1 0]
                                [1 1 0]
                                [0 0 0]]
                                [[0 0 0]
                                [0 0 0]
                                [0 0 0]]],
                                dtype=np.bool_
                )

                figure = sly.Mask3D(mask)
                figure_json = figure.to_json()
                json.dumps(figure_json, indent=4)
                # Output: {
                #    "mask_3d": {
                #        "data": "eJzrDPBz5+WS4mJgYOD19HAJAtLMIMwIInOeqf8BUmwBPiGuQPr///9Lb86/C2QxlgT5BTM4PLuRBuTwebo4hlTMSa44sKHhISMDuxpTYrr03F6gDIOnq5/LOqeEJgDM5ht6",
                #    },
                #    "shape": "mask_3d",
                #    "geometryType": "mask_3d"
                # }
        """
        res = {
            self._impl_json_class_name(): {
                DATA: self.data_2_base64(self.data),
            },
            GEOMETRY_SHAPE: self.name(),
            GEOMETRY_TYPE: self.name(),
        }

        if self.space_origin:
            res[f"{self._impl_json_class_name()}"][f"{SPACE_ORIGIN}"] = self.space_origin

        if self.space:
            res[f"{self._impl_json_class_name()}"][f"{SPACE}"] = self.space

        if self.space_directions:
            res[f"{self._impl_json_class_name()}"][f"{SPACE_DIRECTIONS}"] = self.space_directions

        self._add_creation_info(res)
        return res

    @classmethod
    def from_json(cls, json_data: Dict) -> Mask3D:
        """
        Convert a json dict to Mask 3D.

        :param json_data: Mask in json format as a dict.
        :type json_data: Dict
        :returns: Mask3D from json.
        :rtype: :class:`~supervisely.geometry.mask_3d.Mask3D`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                figure_json = {
                    "mask_3d": {
                        "data": "eJzrDPBz5+WS4mJgYOD19HAJAtLMIMwIInOeqf8BUmwBPiGuQPr///9Lb86/C2QxlgT5BTM4PLuRBuTwebo4hlTMSa44sKHhISMDuxpTYrr03F6gDIOnq5/LOqeEJgDM5ht6",
                    },
                    "shape": "mask_3d",
                    "geometryType": "mask_3d"
                }
                figure = sly.Mask3D.from_json(figure_json)
        """
        if json_data == {}:
            return cls(data=np.zeros((3, 3, 3), dtype=np.bool_))

        json_root_key = cls._impl_json_class_name()
        if json_root_key not in json_data:
            raise ValueError(
                "Data must contain {} field to create Mask3D object.".format(json_root_key)
            )

        if DATA not in json_data[json_root_key]:
            raise ValueError(
                "{} field must contain {} field to create Mask3D object.".format(
                    json_root_key, DATA
                )
            )

        data = cls.base64_2_data(json_data[json_root_key][DATA])

        labeler_login = json_data.get(LABELER_LOGIN, None)
        updated_at = json_data.get(UPDATED_AT, None)
        created_at = json_data.get(CREATED_AT, None)
        sly_id = json_data.get(ID, None)
        class_id = json_data.get(CLASS_ID, None)

        header = {}

        space_origin = json_data[json_root_key].get(SPACE_ORIGIN, None)
        if space_origin is not None:
            header["space origin"] = space_origin

        space = json_data[json_root_key].get(SPACE, None)
        if space is not None:
            header["space"] = space

        space_directions = json_data[json_root_key].get(SPACE_DIRECTIONS, None)
        if space_directions is not None:
            header["space directions"] = space_directions

        return cls(
            data=data.astype(np.bool_),
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
            volume_header=header,
        )

    @classmethod
    def _impl_json_class_name(cls):
        """_impl_json_class_name"""
        return MASK_3D

    @staticmethod
    def data_2_base64(data: np.ndarray) -> str:
        """
        Convert numpy array to base64 encoded string.

        :param data: Bool numpy array.
        :type data: np.ndarray
        :returns: Base64 encoded string
        :rtype: str

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import nrrd
                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                meta_json = api.project.get_meta(PROJECT_ID)
                meta = sly.ProjectMeta.from_json(meta_json)

                ann_json = api.volume.annotation.download_bulk(DATASET_ID, [VOLUME_ID])

                figure_id = ann_json[0]["spatialFigures"][0]["id"]
                path_for_mesh = f"meshes/{figure_id}.nrrd"
                api.volume.figure.download_stl_meshes([figure_id], [path_for_mesh])

                mask3d_data, _ = sly.volume.volume.read_nrrd_serie_volume_np(path_for_mesh)
                encoded_string = sly.Mask3D.data_2_base64(mask3d_data)

                print(encoded_string)
                # 'H4sIAGWoWmQC/zPWMdYxrmFkZAAiIIAz4AAAE56ciyEAAAA='
        """
        shape_str = ",".join(str(dim) for dim in data.shape)
        data_str = data.tostring().decode("utf-8")
        combined_str = f"{shape_str}|{data_str}"
        compressed_string = gzip.compress(combined_str.encode("utf-8"))
        encoded_string = base64.b64encode(compressed_string).decode("utf-8")
        return encoded_string

    @staticmethod
    def base64_2_data(encoded_string: str) -> np.ndarray:
        """
        Convert base64 encoded string to numpy array.

        :param encoded_string: Input base64 encoded string.
        :type encoded_string: str
        :returns: Bool numpy array
        :rtype: np.ndarray

        :Usage Example:

            .. code-block:: python

              import supervisely as sly

              encoded_string = 'H4sIAGWoWmQC/zPWMdYxrmFkZAAiIIAz4AAAE56ciyEAAAA='
              figure_data = sly.Mask3D.base64_2_data(encoded_string)
              print(figure_data)
              # [[[1 1 0]
              #   [1 1 0]
              #   [0 0 0]]
              #  [[1 1 0]
              #   [1 1 0]
              #   [0 0 0]]
              #  [[0 0 0]
              #   [0 0 0]
              #   [0 0 0]]]
        """
        compressed_bytes = base64.b64decode(encoded_string)
        decompressed_string = gzip.decompress(compressed_bytes).decode("utf-8")
        shape_str, data_str = decompressed_string.split("|")
        shape = tuple(int(dim) for dim in shape_str.split(","))
        data_bytes = data_str.encode("utf-8")
        try:
            data = np.frombuffer(data_bytes, dtype=np.uint8).reshape(shape)
        except ValueError:
            logger.warn(
                "Can't reshape array with 'dtype=np.uint8'. Will try to automatically convert 'dtype=np.int16' to 'np.uint8' and reshape"
            )
            data = np.frombuffer(data_bytes, dtype=np.int16)
            data = np.clip(data, 0, 1).astype(np.uint8)
            data = data.reshape(shape)
            logger.debug("Converted successfully!")
        return data

    def add_mask_2d(
        self,
        mask_2d: np.ndarray,
        plane_name: Literal["axial", "sagittal", "coronal"],
        slice_index: int,
        origin: Optional[List[int]] = None,
    ):
        """
        Draw a 2D mask on a 3D Mask.

        :param mask_2d: 2D array with a flat mask.
        :type mask_2d: np.ndarray
        :param plane_name: Name of the plane: "axial", "sagittal", "coronal".
        :type plane_name: str
        :param slice_index: Slice index of the volume figure.
        :type slice_index: int
        :param origin: (row, col) position. The top-left corner of the mask is located on the specified slice (optional).
        :type origin: Optional[List[int]], NoneType
        """

        from supervisely.volume_annotation.plane import Plane

        Plane.validate_name(plane_name)

        mask_2d = np.fliplr(mask_2d)
        mask_2d = np.rot90(mask_2d, 1, (1, 0))

        if plane_name == Plane.AXIAL:
            new_shape = self.data.shape[:2]
        elif plane_name == Plane.SAGITTAL:
            new_shape = self.data.shape[1:]
        elif plane_name == Plane.CORONAL:
            new_shape = self.data.shape[::2]

        if origin:
            x, y = origin
            # pylint: disable=possibly-used-before-assignment
            new_mask = np.zeros(new_shape, dtype=mask_2d.dtype)
            new_mask[x : x + mask_2d.shape[0], y : y + mask_2d.shape[1]] = mask_2d

        if plane_name == Plane.AXIAL:
            self.data[:, :, slice_index] = new_mask
        elif plane_name == Plane.SAGITTAL:
            self.data[slice_index, :, :] = new_mask
        elif plane_name == Plane.CORONAL:
            self.data[:, slice_index, :] = new_mask

    @staticmethod
    def _bytes_from_nrrd(path: str) -> Tuple[str, bytes]:
        """
        Read geometry from a file as bytes.

        The NRRD file must be named with a hexadecimal UUID value. Only NRRD files are supported.

        :param path: Path to the NRRD file containing geometry.
        :type path: str
        :returns: A tuple containing the key hex value and geometry bytes, or (None, None) if the file is not found.
        :rtype: Tuple[str, bytes]
        """

        if get_file_ext(path) == ".nrrd":
            key = get_file_name(path)
            with open(path, "rb") as file:
                geometry_bytes = file.read()
            return key, geometry_bytes
        else:
            return None, None

    @staticmethod
    def _bytes_from_nrrd_batch(paths: List[str]) -> Dict[str, bytes]:
        """
        Read geometries from multiple files as bytes and map them to figure UUID hex values in a dictionary.

        The NRRD files must be named with a hexadecimal UUID value. Only NRRD files are supported.

        :param paths: Paths to the NRRD files containing geometry.
        :type paths: List[str]
        :returns: A dictionary mapping figure UUID hex values to their respective geometries.
        :rtype: Dict[str, bytes]
        """
        geometries_dict = {}
        for path in paths:
            key, geometry_bytes = Mask3D._bytes_from_nrrd(path)
            if key is None and geometry_bytes is None:
                continue
            geometries_dict[key] = geometry_bytes
        return geometries_dict

    def set_volume_space_meta(self, header: Dict):
        """
        Set space, space directions, and space origin attributes from a NRRD header dictionary.

        :param header: NRRD header dictionary.
        :type header: dict
        """
        if "space" in header:
            self.space = header["space"]
        if "space directions" in header:
            self.space_directions = header["space directions"]
        if "space origin" in header:
            self.space_origin = PointVolume(
                x=header["space origin"][0],
                y=header["space origin"][1],
                z=header["space origin"][2],
            )

    def create_header(self) -> OrderedDict:
        """
        Create header for encoding Mask3D to NRRD bytes

        :returns: Header for NRRD file
        :rtype: OrderedDict
        """
        header = OrderedDict()
        if self.space is not None:
            header["space"] = self.space
        if self.space_directions is not None:
            header["space directions"] = self.space_directions
        if self.space_origin is not None:
            header["space origin"] = self.space_origin
        return header

    def orient_ras(self) -> None:
        """
        Transforms the mask data and updates spatial metadata (origin, directions, spacing)
        to align with the RAS coordinate system using SimpleITK.

        :rtype: None
        """
        import SimpleITK as sitk

        from supervisely.volume.volume import _sitk_image_orient_ras

        # Convert bool data to uint8 for SimpleITK compatibility
        data_for_sitk = self.data.astype(np.uint8) if self.data.dtype == np.bool_ else self.data
        sitk_volume = sitk.GetImageFromArray(data_for_sitk)
        if self.space_origin is not None:
            sitk_volume.SetOrigin(self.space_origin)
        if self.space_directions is not None:
            # Convert space directions to spacing and direction
            space_directions = np.array(self.space_directions)
            spacing = np.linalg.norm(space_directions, axis=1)
            direction = space_directions / spacing[:, np.newaxis]
            sitk_volume.SetSpacing(spacing)
            sitk_volume.SetDirection(direction.flatten())

        sitk_volume = _sitk_image_orient_ras(sitk_volume)

        # Extract transformed data and update object
        self.data = sitk.GetArrayFromImage(sitk_volume)
        new_direction = np.array(sitk_volume.GetDirection()).reshape(3, 3)
        new_spacing = np.array(sitk_volume.GetSpacing())
        new_space_directions = (new_direction.T * new_spacing[:, None]).tolist()
        new_header = {
            "space": "right-anterior-superior",
            "space directions": new_space_directions,
            "space origin": sitk_volume.GetOrigin(),
        }
        self.set_volume_space_meta(new_header)



# coding: utf-8

# docs
from __future__ import annotations

from typing import Dict, List, Optional, Tuple

import cv2
import numpy as np

from supervisely.geometry.constants import (
    CLASS_ID,
    CREATED_AT,
    FACES,
    ID,
    LABELER_LOGIN,
    POINTS,
    UPDATED_AT,
)
from supervisely.geometry.geometry import Geometry
from supervisely.geometry.image_rotator import ImageRotator
from supervisely.geometry.point_location import (
    PointLocation,
    points_to_row_col_list,
    row_col_list_to_points,
)
from supervisely.geometry.rectangle import Rectangle

if not hasattr(np, "bool"):
    np.bool = np.bool_


class CuboidFace:
    """
    A single face of a :class:`~supervisely.geometry.cuboid.Cuboid`.

    The face is defined by four integer indices (``a``, ``b``, ``c``, ``d``) referencing cuboid nodes.
    This lightweight helper is used to store faces and serialize/deserialize them in Supervisely
    annotation format (see :meth:`to_json` / :meth:`from_json`).
    """

    def __init__(self, a: int, b: int, c: int, d: int):
        """
        CuboidFace for a single :class:`~supervisely.geometry.cuboid.Cuboid`.

        :param a: Node of the CuboidFace.
        :type a: int
        :param b: Node of the CuboidFace.
        :type b: int
        :param c: Node of the CuboidFace.
        :type c: int
        :param d: Node of the CuboidFace.
        :type d: int

        :Usage Example:

            .. code-block:: python

                edge = CuboidFace(0, 1, 2, 3)
        """
        self._a = a
        self._b = b
        self._c = c
        self._d = d

    def to_json(self) -> List[int]:
        """
        Convert the CuboidFace to list. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :returns: List of integers
        :rtype: :class:`List[int, int, int, int]`

        :Usage Example:

            .. code-block:: python

                edge = CuboidFace(0, 1, 2, 3)
                edge_json = edge.to_json()
                print(edge_json)
                # Output: [0, 1, 2, 3]
        """
        return [self.a, self.b, self.c, self.d]

    @classmethod
    def from_json(cls, data: List[int]) -> CuboidFace:
        """
        Convert list of integers to CuboidFace. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: List of integers.
        :type data: List[int, int, int, int]
        :raises ValueError: if data have not 4 indices
        :returns: CuboidFace from json.
        :rtype: :class:`~supervisely.geometry.cuboid.CuboidFace`

        :Usage Example:

            .. code-block:: python

                new_edge = CuboidFace.from_json([0, 1, 2, 3])
        """
        if len(data) != 4:
            raise ValueError(
                f"CuboidFace JSON data must have exactly 4 indices, instead got {len(data)!r}."
            )
        return cls(data[0], data[1], data[2], data[3])

    @property
    def a(self):
        """ """
        return self._a

    @property
    def b(self):
        """ """
        return self._b

    @property
    def c(self):
        """ """
        return self._c

    @property
    def d(self):
        """ """
        return self._d

    def tolist(self) -> List[int]:
        """
        Convert CuboidFace to list.

        :returns: List of integers.
        :rtype: :class:`List[int, int, int, int]`

        :Usage Example:

            .. code-block:: python

                edge = CuboidFace(0, 1, 2, 3)
                print(edge.tolist())
                # Output: [0, 1, 2, 3]
        """
        return [self.a, self.b, self.c, self.d]


class Cuboid(Geometry):
    """3D cuboid defined by eight corner points and six faces. Immutable."""

    @staticmethod
    def geometry_name():
        """
        Returns the name of the geometry.

        :returns: name of the geometry
        :rtype: str
        """
        return "cuboid"

    def __init__(
        self,
        points: List[PointLocation],
        faces: List[CuboidFace],
        sly_id: Optional[int] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[int] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
    ):
        """
        Cuboid geometry for a single :class:`~supervisely.annotation.label.Label`. :class:`~supervisely.geometry.cuboid.Cuboid` class object is immutable.

        :param points: List or tuple of :class:`~supervisely.geometry.point_location.PointLocation` objects.
        :type points: List[:class:`~supervisely.geometry.point_location.PointLocation`] or Tuple[:class:`~supervisely.geometry.point_location.PointLocation`]
        :param faces: List or tuple of :class:`~supervisely.geometry.cuboid.CuboidFace` objects.
        :type faces: List[:class:`~supervisely.geometry.cuboid.CuboidFace`] or Tuple[:class:`~supervisely.geometry.cuboid.CuboidFace`]
        :param sly_id: Cuboid ID in Supervisely server.
        :type sly_id: int, optional
        :param class_id: ID for ObjClass to which belongs Cuboid.
        :type class_id: int, optional
        :param labeler_login: Login of the user who created :class:`~supervisely.geometry.cuboid.Cuboid`.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when Cuboid was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when Cuboid was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional
        :raises ValueError: if len(:class:`~supervisely.geometry.cuboid.CuboidFace` objects) != 3. A cuboid must have exactly 3 faces. Instead got {len(faces)} faces.
        :raises ValueError: if point index is out of bounds for cuboid face. Got {len(points)!r} points, but the index is {point_idx!r}.

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.cuboid import CuboidFace

                nodes = [[277, 273], [840, 273], [840, 690], [277, 690], [688, 168], [1200, 168], [1200, 522]]
                edges = [CuboidFace(0, 1, 2, 3), CuboidFace(0, 4, 5, 1), CuboidFace(1, 5, 6, 2)]
                pl_nodes = (sly.PointLocation(node[0], node[1]) for node in nodes)
                figure = sly.Cuboid(pl_nodes, edges)
        """

        super().__init__(
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

        points = list(points)
        faces = list(faces)

        if len(faces) != 3:
            raise ValueError(f"A cuboid must have exactly 3 faces. Instead got {len(faces)} faces.")

        for face in faces:
            for point_idx in (face.a, face.b, face.c, face.d):
                if point_idx >= len(points) or point_idx < 0:
                    raise ValueError(
                        f"Point index is out of bounds for cuboid face. Got {len(points)!r} points, but "
                        f"the index is {point_idx!r}."
                    )

        self._points = points
        self._faces = faces

    @property
    def points(self) -> List[PointLocation]:
        """
        List of :class:`~supervisely.geometry.point_location.PointLocation` objects.

        :returns: Cuboid nodes
        :rtype: List[:class:`~supervisely.geometry.point_location.PointLocation`]
        """
        return self._points.copy()

    @property
    def faces(self) -> List[CuboidFace]:
        """
        List of :class:`~supervisely.geometry.cuboid.CuboidFace` objects.

        :returns: Cuboid edges
        :rtype: List[:class:`~supervisely.geometry.cuboid.CuboidFace`]
        """
        return self._faces.copy()

    def to_json(self) -> Dict:
        """
        Convert the Cuboid to a json dict. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :returns: Json format as a dict
        :rtype: dict

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.cuboid import CuboidFace

                nodes = [[277, 273], [840, 273], [840, 690], [277, 690], [688, 168], [1200, 168], [1200, 522]]
                edges = [CuboidFace(0, 1, 2, 3), CuboidFace(0, 4, 5, 1), CuboidFace(1, 5, 6, 2)]
                pl_nodes = (sly.PointLocation(node[0], node[1]) for node in nodes)
                figure = sly.Cuboid(pl_nodes, edges)

                figure_json = figure.to_json()
                print(figure_json)
                # Output: {
                #    "faces": [
                #        [0, 1, 2, 3],
                #        [0, 4, 5, 1],
                #        [1, 5, 6, 2]
                #    ],
                #    "points": [
                #        [273, 277],
                #        [273, 840],
                #        [690, 840],
                #        [690, 277],
                #        [168, 688],
                #        [168, 1200],
                #        [522, 1200]
                #    ]
                # }
        """
        packed_obj = {
            POINTS: points_to_row_col_list(self._points, flip_row_col_order=True),
            FACES: [face.to_json() for face in self._faces],
        }
        self._add_creation_info(packed_obj)
        return packed_obj

    @classmethod
    def from_json(cls, data: Dict) -> Cuboid:
        """
        Convert a json dict to Cuboid. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: Cuboid in json format as a dict.
        :type data: Dict[str, List[int]]
        :returns: Cuboid from json.
        :rtype: :class:`~supervisely.geometry.cuboid.Cuboid`
        :raises ValueError: if json format is not correct

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                data = {
                    "faces": [
                                [0, 1, 2, 3],
                                [0, 4, 5, 1],
                                [1, 5, 6, 2]
                    ],
                    "points": [
                                [273, 277],
                                [273, 840],
                                [690, 840],
                                [690, 277],
                                [168, 688],
                                [168, 1200],
                                [522, 1200]
                    ]
                }

                figure = sly.Cuboid.from_json(data)
        """
        for k in [POINTS, FACES]:
            if k not in data:
                raise ValueError(f"Field {k!r} not found in Cuboid JSON data.")

        points = row_col_list_to_points(data[POINTS], flip_row_col_order=True)
        faces = [CuboidFace.from_json(face_json) for face_json in data[FACES]]

        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        return cls(
            points=points,
            faces=faces,
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

    def crop(self, rect: Rectangle) -> List[Cuboid]:
        """
        Crops current Cuboid.

        :param rect: Rectangle for crop.
        :type rect: :class:`~supervisely.geometry.rectangle.Rectangle`
        :returns: List of Cuboids after crop.
        :rtype: List[:class:`~supervisely.geometry.cuboid.Cuboid`]

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.cuboid import CuboidFace

                nodes = [[277, 273], [840, 273], [840, 690], [277, 690], [688, 168], [1200, 168], [1200, 522]]
                edges = [CuboidFace(0, 1, 2, 3), CuboidFace(0, 4, 5, 1), CuboidFace(1, 5, 6, 2)]
                pl_nodes = (sly.PointLocation(node[0], node[1]) for node in nodes)
                figure = sly.Cuboid(pl_nodes, edges)
                crop_figures = figure.crop(sly.Rectangle(1, 1, 1500, 1550))
        """
        is_all_nodes_inside = all(
            rect.contains_point_location(self._points[p])
            for face in self._faces
            for p in face.tolist()
        )
        return [self] if is_all_nodes_inside else []

    def _transform(self, transform_fn):
        """ """
        return Cuboid(points=[transform_fn(p) for p in self.points], faces=self.faces)

    def rotate(self, rotator: ImageRotator) -> Cuboid:
        """
        Rotates current Cuboid.

        :param rotator: Class for image rotation.
        :type rotator: :class:`~supervisely.geometry.image_rotator.ImageRotator`
        :returns: Cuboid after rotation.
        :rtype: :class:`~supervisely.geometry.cuboid.Cuboid`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.cuboid import CuboidFace
                from supervisely.geometry.image_rotator import ImageRotator

                nodes = [[277, 273], [840, 273], [840, 690], [277, 690], [688, 168], [1200, 168], [1200, 522]]
                edges = [CuboidFace(0, 1, 2, 3), CuboidFace(0, 4, 5, 1), CuboidFace(1, 5, 6, 2)]
                pl_nodes = (sly.PointLocation(node[0], node[1]) for node in nodes)
                figure = sly.Cuboid(pl_nodes, edges)

                # Remember that Cuboid class object is immutable, and we need to assign new instance of Cuboid to a new variable
                rotator = ImageRotator((300, 400), 25)
                rotate_fugure= figure.rotate(rotator)
        """
        return self._transform(lambda p: rotator.transform_point(p))

    def resize(self, in_size: Tuple[int, int], out_size: Tuple[int, int]) -> Cuboid:
        """
        Resize current Cuboid.

        :param in_size: Input image size (height, width) to which belongs :class:`~supervisely.geometry.cuboid.Cuboid` object.
        :type in_size: Tuple[int, int]
        :param out_size: Desired output image size (height, width) to which belongs :class:`~supervisely.geometry.cuboid.Cuboid` object.
        :type out_size: Tuple[int, int]
        :returns: Cuboid after resize.
        :rtype: :class:`~supervisely.geometry.cuboid.Cuboid`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.cuboid import CuboidFace

                nodes = [[277, 273], [840, 273], [840, 690], [277, 690], [688, 168], [1200, 168], [1200, 522]]
                edges = [CuboidFace(0, 1, 2, 3), CuboidFace(0, 4, 5, 1), CuboidFace(1, 5, 6, 2)]
                pl_nodes = (sly.PointLocation(node[0], node[1]) for node in nodes)
                figure = sly.Cuboid(pl_nodes, edges)

                # Remember that Cuboid class object is immutable, and we need to assign new instance of Cuboid to a new variable
                in_height, in_width = 1250, 1400
                out_height, out_width = 600, 800
                resize_figure = figure.resize((in_height, in_width), (out_height, out_width))
        """
        return self._transform(lambda p: p.resize(in_size, out_size))

    def scale(self, factor: float) -> Cuboid:
        """
        Scale current Cuboid.

        :param factor: Scale parameter.
        :type factor: float
        :returns: Cuboid after scale.
        :rtype: :class:`~supervisely.geometry.cuboid.Cuboid`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.cuboid import CuboidFace

                nodes = [[277, 273], [840, 273], [840, 690], [277, 690], [688, 168], [1200, 168], [1200, 522]]
                edges = [CuboidFace(0, 1, 2, 3), CuboidFace(0, 4, 5, 1), CuboidFace(1, 5, 6, 2)]
                pl_nodes = (sly.PointLocation(node[0], node[1]) for node in nodes)
                figure = sly.Cuboid(pl_nodes, edges)

                # Remember that Cuboid class object is immutable, and we need to assign new instance of Cuboid to a new variable
                scale_figure = figure.scale(1.75)
        """
        return self._transform(lambda p: p.scale(factor))

    def translate(self, drow: int, dcol: int) -> Cuboid:
        """
        Translate current Cuboid.

        :param drow: Horizontal shift.
        :type drow: int
        :param dcol: Vertical shift.
        :type dcol: int
        :returns: Cuboid after translate.
        :rtype: :class:`~supervisely.geometry.cuboid.Cuboid`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.cuboid import CuboidFace

                nodes = [[277, 273], [840, 273], [840, 690], [277, 690], [688, 168], [1200, 168], [1200, 522]]
                edges = [CuboidFace(0, 1, 2, 3), CuboidFace(0, 4, 5, 1), CuboidFace(1, 5, 6, 2)]
                pl_nodes = (sly.PointLocation(node[0], node[1]) for node in nodes)
                figure = sly.Cuboid(pl_nodes, edges)

                # Remember that Cuboid class object is immutable, and we need to assign new instance of Cuboid to a new variable
                translate_figure = figure.translate(150, 350)
        """
        return self._transform(lambda p: p.translate(drow, dcol))

    def fliplr(self, img_size: Tuple[int, int]) -> Cuboid:
        """
        Flips current Cuboid in horizontal.

        :param img_size: Image size (height, width) to which belongs :class:`~supervisely.geometry.cuboid.Cuboid` object.
        :type img_size: Tuple[int, int]
        :returns: Cuboid after flip in horizontal.
        :rtype: :class:`~supervisely.geometry.cuboid.Cuboid`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.cuboid import CuboidFace

                nodes = [[277, 273], [840, 273], [840, 690], [277, 690], [688, 168], [1200, 168], [1200, 522]]
                edges = [CuboidFace(0, 1, 2, 3), CuboidFace(0, 4, 5, 1), CuboidFace(1, 5, 6, 2)]
                pl_nodes = (sly.PointLocation(node[0], node[1]) for node in nodes)
                figure = sly.Cuboid(pl_nodes, edges)

                # Remember that Cuboid class object is immutable, and we need to assign new instance of Cuboid to a new variable
                height, width = 1250, 1400
                fliplr_figure = figure.fliplr((height, width))
        """
        return self._transform(lambda p: p.fliplr(img_size))

    def flipud(self, img_size: Tuple[int, int]) -> Cuboid:
        """
        Flips current Cuboid in vertical.

        :param img_size: Image size (height, width) to which belongs :class:`~supervisely.geometry.cuboid.Cuboid` object.
        :type img_size: Tuple[int, int]
        :returns: Cuboid after flip in vertical.
        :rtype: :class:`~supervisely.geometry.cuboid.Cuboid`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.cuboid import CuboidFace

                nodes = [[277, 273], [840, 273], [840, 690], [277, 690], [688, 168], [1200, 168], [1200, 522]]
                edges = [CuboidFace(0, 1, 2, 3), CuboidFace(0, 4, 5, 1), CuboidFace(1, 5, 6, 2)]
                pl_nodes = (sly.PointLocation(node[0], node[1]) for node in nodes)
                figure = sly.Cuboid(pl_nodes, edges)

                # Remember that Cuboid class object is immutable, and we need to assign new instance of Cuboid to a new variable
                height, width = 1250, 1400
                flipud_figure = figure.flipud((height, width))
        """
        return self._transform(lambda p: p.flipud(img_size))

    def _draw_impl(self, bitmap: np.ndarray, color, thickness=1, config=None):
        """ """
        bmp_to_draw = np.zeros(bitmap.shape[:2], np.uint8)
        for contour in self._contours_list():
            cv2.fillPoly(bmp_to_draw, pts=[np.array(contour, dtype=np.int32)], color=1)
        bool_mask = bmp_to_draw.astype(bool)
        bitmap[bool_mask] = color

    def _draw_contour_impl(self, bitmap, color, thickness=1, config=None):
        """ """
        contours_np_list = [np.array(contour, dtype=np.int32) for contour in self._contours_list()]
        cv2.polylines(bitmap, pts=contours_np_list, isClosed=True, color=color, thickness=thickness)

    def _contours_list(self):
        """ """
        return [
            points_to_row_col_list(
                [self._points[idx] for idx in face.tolist()], flip_row_col_order=True
            )
            for face in self._faces
        ]

    def to_bbox(self) -> Rectangle:
        """
        Create Rectangle object from current Cuboid.

        :returns: Rectangle from cuboid.
        :rtype: :class:`~supervisely.geometry.rectangle.Rectangle`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.cuboid import CuboidFace

                nodes = [[277, 273], [840, 273], [840, 690], [277, 690], [688, 168], [1200, 168], [1200, 522]]
                edges = [CuboidFace(0, 1, 2, 3), CuboidFace(0, 4, 5, 1), CuboidFace(1, 5, 6, 2)]
                pl_nodes = (sly.PointLocation(node[0], node[1]) for node in nodes)
                figure = sly.Cuboid(pl_nodes, edges)

                rectangle = figure.to_bbox()
        """
        points_np = np.array(
            [
                [self._points[p].row, self._points[p].col]
                for face in self._faces
                for p in face.tolist()
            ]
        )
        rows, cols = points_np[:, 0], points_np[:, 1]
        return Rectangle(
            top=round(min(rows).item()),
            left=round(min(cols).item()),
            bottom=round(max(rows).item()),
            right=round(max(cols).item()),
        )

    @property
    def area(self) -> float:
        """
        Cuboid area.

        :returns: Area of current Cuboid.
        :rtype: float

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.cuboid import CuboidFace

                nodes = [[277, 273], [840, 273], [840, 690], [277, 690], [688, 168], [1200, 168], [1200, 522]]
                edges = [CuboidFace(0, 1, 2, 3), CuboidFace(0, 4, 5, 1), CuboidFace(1, 5, 6, 2)]
                pl_nodes = (sly.PointLocation(node[0], node[1]) for node in nodes)
                figure = sly.Cuboid(pl_nodes, edges)

                print(figure.area)
                # Output: 5146.0
        """
        bbox = self.to_bbox()
        canvas = np.zeros([bbox.bottom + 1, bbox.right + 1], dtype=np.bool)
        self.draw(canvas, True)
        return float(np.sum(canvas))



import json
from pathlib import Path
from typing import Optional

from jinja2 import Template

from supervisely.api.api import Api
from supervisely.io.fs import dir_empty, get_directory_size
from supervisely.nn.benchmark.visualization.widgets import BaseWidget
from supervisely.task.progress import tqdm_sly
from supervisely import logger


class Renderer:
    """
    Base class for rendering visualizations of Evaluation Report.
    """

    def __init__(
        self,
        layout: BaseWidget,
        base_dir: str = "./output",
        template: str = None,
        report_name: str = "Model Evaluation Report.lnk",
    ) -> None:
        """:param layout: Root widget layout.
        :param base_dir: Output directory (must be empty).
        :param template: HTML template string. Uses default if None.
        :param report_name: Report link filename.
        """
        if template is None:
            template = (
                Path(__file__).parents[1].joinpath("visualization/report_template.html").read_text()
            )
        self.main_template = template
        self.layout = layout
        self.base_dir = base_dir
        self.report_name = report_name
        self._report = None
        self._lnk = None

        if Path(base_dir).exists():
            if not dir_empty(base_dir):
                raise ValueError(f"Output directory {base_dir} is not empty.")

    @property
    def _template_data(self):
        return {"layout": self.layout.to_html()}

    def render(self):
        return Template(self.main_template).render(self._template_data)

    def get_state(self):
        return {}

    def save(self) -> None:
        self.layout.save_data(self.base_dir)
        state = self.layout.get_state()
        with open(Path(self.base_dir).joinpath("state.json"), "w") as f:
            json.dump(state, f)
        template = self.render()
        with open(Path(self.base_dir).joinpath("template.vue"), "w") as f:
            f.write(template)
        return template

    def visualize(self):
        return self.save()

    def upload_results(
        self, api: Api, team_id: int, remote_dir: str, progress: Optional[tqdm_sly] = None
    ) -> str:
        if dir_empty(self.base_dir):
            raise RuntimeError(
                "No visualizations to upload. You should call visualize method first."
            )
        if progress is None:
            progress = tqdm_sly
        dir_total = get_directory_size(self.base_dir)
        dir_name = Path(remote_dir).name
        with progress(
            message=f"Uploading visualizations to {dir_name}",
            total=dir_total,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
        ) as pbar:
            remote_dir = api.file.upload_directory(
                team_id,
                self.base_dir,
                remote_dir,
                change_name_if_conflict=True,
                progress_size_cb=pbar,
            )
        src = self._save_report_link(api, team_id, remote_dir)
        dst = Path(remote_dir).joinpath(self.report_name)
        self._lnk = api.file.upload(team_id=team_id, src=src, dst=str(dst))
        return remote_dir

    def _save_report_link(self, api: Api, team_id: int, remote_dir: str):
        report_link = self._get_report_path(api, team_id, remote_dir)
        pth = Path(self.base_dir).joinpath(self.report_name)
        with open(pth, "w") as f:
            f.write(report_link)
        logger.debug(f"Report link: {self._get_report_link(api, team_id, remote_dir)}")
        return str(pth)

    def _get_report_link(self, api: Api, team_id: int, remote_dir: str):
        path = self._get_report_path(api, team_id, remote_dir)
        return f"{api.server_address}{path}"

    def _get_report_path(self, api: Api, team_id: int, remote_dir: str):
        template_path = Path(remote_dir).joinpath("template.vue")
        self._report = api.file.get_info_by_path(team_id, str(template_path))
        return "/model-benchmark?id=" + str(self._report.id)

    @property
    def report(self):
        return self._report

    @property
    def lnk(self):
        return self._lnk



# coding: utf-8

import uuid

from supervisely._utils import take_with_default
from supervisely.annotation.tag import Tag, TagJsonFields
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.volume_annotation.constants import ID, KEY


class VolumeTag(Tag):
    """Tag on a volume annotation (meta, value). Immutable."""

    def __init__(
        self,
        meta,
        value=None,
        key=None,
        sly_id=None,
        labeler_login=None,
        updated_at=None,
        created_at=None,
    ):
        """
        Tag on volume annotation.

        :param meta: Tag metadata (name, value type).
        :type meta: :class:`~supervisely.annotation.tag_meta.TagMeta`
        :param value: Tag value; type must match TagMeta.value_type.
        :type value: str or int or float, optional
        :param key: UUID key. Auto-generated if not provided.
        :type key: uuid.UUID, optional
        :param sly_id: Server-side tag ID.
        :type sly_id: int, optional
        :param labeler_login: Login of user who created the tag.
        :type labeler_login: str, optional
        :param updated_at: Last modification timestamp (ISO format).
        :type updated_at: str, optional
        :param created_at: Creation timestamp (ISO format).
        :type created_at: str, optional

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta_brain = sly.TagMeta('brain', sly.TagValueType.NONE)
                tag_brain = sly.VolumeTag(meta_brain)

                meta_heart = sly.TagMeta('heart', sly.TagValueType.ANY_STRING)
                tag_heart = sly.VolumeTag(meta_heart, value="Heart")
        """
        super(VolumeTag, self).__init__(
            meta,
            value=value,
            sly_id=sly_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )
        self._key = take_with_default(key, uuid.uuid4())

    def key(self) -> str:
        """
        Get key value.

        :returns: Get key value.
        :rtype: str

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.volume_annotation.volume_tag import VolumeTag

                meta_heart = sly.TagMeta('heart', sly.TagValueType.NONE)
                tag_heart = VolumeTag(meta_heart)

                print(tag_heart.key())
                # Output: 2d9cd2cd-f89c-40a9-8675-1b0773fa250d
        """
        return self._key

    def to_json(self, key_id_map: KeyIdMap = None):
        """
        Convert the VolumeTag to a json dict. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param key_id_map: Key ID Map object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :returns: Json format as a dict
        :rtype: dict

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                from supervisely.volume_annotation.volume_tag import VolumeTag
                meta_heart = sly.TagMeta('heart', sly.TagValueType.NONE)
                tag_heart = VolumeTag(meta_heart)

                tag_heart_json = tag_heart.to_json()

                print(tag_heart_json)
                # Output: {
                #     "name": "heart",
                #     "key": "058ad7993a534082b4d94cc52542a97d"
                # }
        """

        data_json = super(VolumeTag, self).to_json()
        if type(data_json) is str:
            # @TODO: case when tag has no value, super.to_json() returns tag name
            data_json = {TagJsonFields.TAG_NAME: data_json}
        data_json[KEY] = self.key().hex

        if key_id_map is not None:
            item_id = key_id_map.get_tag_id(self.key())
            if item_id is not None:
                data_json[ID] = item_id

        return data_json

    @classmethod
    def from_json(cls, data: dict, tag_meta_collection, key_id_map: KeyIdMap = None):
        """
        Convert a json dict to VolumeTag. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: VolumeTag in json format as a dict.
        :type data: dict
        :param tag_meta_collection: TagMetaCollection object.
        :type tag_meta_collection: :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection`
        :param key_id_map: Key ID Map object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :returns: VolumeTag object
        :rtype: :class:`~supervisely.volume_annotation.volume_tag.VolumeTag`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.volumme_annotation.volume_tag import VolumeTag

                tag_heart_json = {"name": "heart", "value": "777"}
                meta_heart = sly.TagMeta('heart', sly.TagValueType.ANY_STRING)
                meta_collection = sly.TagMetaCollection([meta_heart])

                tag_heart = VolumeTag.from_json(tag_heart_json, meta_collection)
        """

        temp = super(VolumeTag, cls).from_json(data, tag_meta_collection)
        key = uuid.UUID(data[KEY]) if KEY in data else uuid.uuid4()

        if key_id_map is not None:
            key_id_map.add_tag(key, data.get(ID, None))

        return cls(
            meta=temp.meta,
            value=temp.value,
            key=key,
            sly_id=temp.sly_id,
            labeler_login=temp.labeler_login,
            updated_at=temp.updated_at,
            created_at=temp.created_at,
        )

    def clone(
        self,
        meta=None,
        value=None,
        key=None,
        sly_id=None,
        labeler_login=None,
        updated_at=None,
        created_at=None,
    ):
        """
        Makes a copy of VolumeTag with new fields, if fields are given, otherwise it will use fields of the original VolumeTag.

        :param meta: General information about VolumeTag.
        :type meta: :class:`~supervisely.annotation.tag_meta.TagMeta`, optional
        :param value: VolumeTag value. Depends on TagValueType of TagMeta.
        :type value: str or int or float or None, optional
        :param key: The UUID key associated with the :class:`~supervisely.volume_annotation.volume_figure.VolumeFigure`.
        :type key: uuid.UUID, optional
        :param sly_id: VolumeTag ID in Supervisely.
        :type sly_id: int, optional
        :param labeler_login: Login of user who created VolumeTag.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when VolumeTag was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when VolumeTag was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.volume_annotation.volume_tag import VolumeTag

                meta_heart = sly.TagMeta('heart_tag', sly.TagValueType.ANY_STRING)
                heart_tag = VolumeTag(meta_heart, value='Heart')

                meta_heart_2 = sly.TagMeta('heart tag 2', sly.TagValueType.ANY_STRING)

                new_tag = car_tag.clone(meta=meta_heart_2, key=car_tag.key())
                new_tag_json = new_tag.to_json()

                print(new_tag_json)
                # Output: {
                #     "name": "heart tag 2",
                #     "value": "Heart",
                #     "key": "360438485fd34264921ca19bd43b0b71"
                # }
        """

        return VolumeTag(
            meta=take_with_default(meta, self.meta),
            value=take_with_default(value, self.value),
            key=take_with_default(key, self.key),
            sly_id=take_with_default(sly_id, self.sly_id),
            labeler_login=take_with_default(labeler_login, self.labeler_login),
            updated_at=take_with_default(updated_at, self.updated_at),
            created_at=take_with_default(created_at, self.created_at),
        )



from pathlib import Path
import cv2
import numpy as np
from .osnet import osnet_x1_0
from collections import OrderedDict
from supervisely import logger

try:
    # pylint: disable=import-error
    import torch
    from torch.nn import functional as F
except ImportError:
    logger.warning("torch is not installed, OSNet re-ID cannot be used.")


class OsnetReIDModel:
    """OSNet re-identification model wrapper that extracts normalized appearance embeddings for detections."""

    def __init__(self, weights_path: Path = None, device: torch.device = torch.device("cpu"), half: bool = False):
        """
        :param weights_path: Path to weights.
        :type weights_path: Path
        :param device: Torch device.
        :type device: torch.device
        :param half: FP16.
        :type half: bool
        """
        self.device = device
        self.half = half
        self.input_shape = (256, 128)
        if weights_path is None:
            self.model = osnet_x1_0(num_classes=1000, loss='softmax', pretrained=True, use_gpu=device)
        else:
            self.model = osnet_x1_0(num_classes=1000, loss='softmax', pretrained=False, use_gpu=device)
            self.load_pretrained_weights(weights_path)
        self.model.to(self.device).eval()
        if self.half:
            self.model.half()

    def load_pretrained_weights(self, weight_path: Path):
        checkpoint = torch.load(weight_path, map_location=self.device)
        state_dict = checkpoint.get("state_dict", checkpoint)
        model_dict = self.model.state_dict()

        new_state_dict = OrderedDict()
        for k, v in state_dict.items():
            key = k[7:] if k.startswith("module.") else k
            if key in model_dict and model_dict[key].size() == v.size():
                new_state_dict[key] = v

        model_dict.update(new_state_dict)
        self.model.load_state_dict(model_dict)

    def get_features(self, xyxys, img: np.ndarray):
        if xyxys.size == 0:
            return np.empty((0, 512))

        crops = self._get_crops(xyxys, img)
        with torch.no_grad():
            features = self.model(crops)
            features = F.normalize(features, dim=1).cpu().numpy()

        return features

    def _get_crops(self, xyxys, img):
        h, w = img.shape[:2]
        crops = []

        for box in xyxys:
            x1, y1, x2, y2 = box.round().astype(int)
            x1, y1, x2, y2 = max(0, x1), max(0, y1), min(w, x2), min(h, y2)
            crop = cv2.resize(img[y1:y2, x1:x2], self.input_shape[::-1])
            crop = cv2.cvtColor(crop, cv2.COLOR_BGR2RGB)
            crop = crop.astype(np.float32) / 255.0
            crop = (crop - [0.485, 0.456, 0.406]) / [0.229, 0.224, 0.225]
            crop = torch.from_numpy(crop).permute(2, 0, 1)
            crops.append(crop)

        batch = torch.stack(crops).to(self.device, dtype=torch.float16 if self.half else torch.float32)
        return batch


class OsnetReIDInterface:
    """Small facade around :class:`OsnetReIDModel` used by trackers to compute embeddings for detections."""

    def __init__(self, weights: Path, device: str = "cpu", fp16: bool = False):
        """
        :param weights: Path to weights.
        :type weights: Path
        :param device: Device string.
        :type device: str
        :param fp16: Use FP16.
        :type fp16: bool
        """
        self.device = torch.device(device)
        self.fp16 = fp16
        self.model = OsnetReIDModel(weights, self.device, half=fp16)

    def inference(self, image: np.ndarray, detections: np.ndarray) -> np.ndarray:
        if detections is None or np.size(detections) == 0:
            return np.zeros((0, 512), dtype=np.float32)  # пустой набор фичей

        xyxys = detections[:, 0:4]  # left, top, right, bottom
        features = self.model.get_features(xyxys, image)
        return features



# coding: utf-8

# docs
from __future__ import annotations

from typing import Any, Dict, List, Literal, Optional, Tuple, Union

from supervisely._utils import take_with_default
from supervisely.annotation.obj_class import ObjClass
from supervisely.annotation.obj_class_collection import ObjClassCollection
from supervisely.annotation.tag_meta import TagMeta
from supervisely.annotation.tag_meta_collection import TagMetaCollection
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.rectangle import Rectangle
from supervisely.io.json import JsonSerializable
from supervisely.project.project_settings import LabelingInterface, ProjectSettings
from supervisely.project.project_type import ProjectType


class ProjectMetaJsonFields:
    """
    JSON field names used in :class:`~supervisely.project.project_meta.ProjectMeta` serialization.
    """
    OBJ_CLASSES = "classes"
    IMG_TAGS = "tags_images"
    OBJ_TAGS = "tags_objects"
    TAGS = "tags"
    PROJECT_TYPE = "projectType"
    PROJECT_SETTINGS = "projectSettings"


def _merge_img_obj_tag_metas(
    img_tag_metas: ObjClassCollection, obj_tag_metas: ObjClassCollection
) -> ObjClassCollection:
    obj_tag_metas_to_add = []
    for obj_tag_meta in obj_tag_metas:
        img_tag_meta_same_key = img_tag_metas.get(obj_tag_meta.key(), None)
        if img_tag_meta_same_key is None:
            obj_tag_metas_to_add.append(obj_tag_meta)
        elif not img_tag_meta_same_key.is_compatible(obj_tag_meta):
            raise ValueError(
                "Unable to merge tag metas for images and objects. Found tags with the same name, but incompatible "
                "values. \n Image-level tag meta: {}\n Object-level tag meta: {}.\n Rename one of the tags to have a "
                "unique name to be able to load project meta.".format(
                    str(img_tag_meta_same_key), str(obj_tag_meta)
                )
            )
    return img_tag_metas.add_items(obj_tag_metas_to_add)


class ProjectMeta(JsonSerializable):
    """
    Project-level metadata: object classes, tag metas, project type and settings.

    This schema is used to validate and interpret annotations across a Supervisely project.
    """

    def __init__(
        self,
        obj_classes: Optional[Union[ObjClassCollection, List[ObjClass]]] = None,
        tag_metas: Optional[Union[TagMetaCollection, List[TagMeta]]] = None,
        project_type: Optional[ProjectType] = None,
        project_settings: Optional[Union[ProjectSettings, Dict]] = None,
    ):
        """
        General information about ProjectMeta. ProjectMeta object is immutable.

        :param obj_classes: ObjClassCollection or just list that stores :class:`~supervisely.annotation.obj_class.ObjClass` instances with unique names.
        :type obj_classes: Union[:class:`~supervisely.annotation.obj_class_collection.ObjClassCollection`, List[:class:`~supervisely.annotation.obj_class.ObjClass`]], optional
        :param tag_metas: TagMetaCollection or just list that stores :class:`~supervisely.annotation.tag_meta.TagMeta` instances with unique names.
        :type tag_metas: Union[:class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection`, List[:class:`~supervisely.annotation.tag_meta.TagMeta`]], optional
        :param project_type: Type of items in project: images, videos, volumes, point_clouds.
        :type project_type: :class:`~supervisely.project.project_type.ProjectType`, optional
        :param project_settings: Additional project properties. For example, multi-view settings.
        :type project_settings: Union[:class:`~supervisely.project.project_settings.ProjectSettings`, Dict[str, Any]], optional

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                # Example 1: Empty ProjectMeta
                meta = sly.ProjectMeta()
                print(meta)
                # Output:
                # ProjectMeta:
                # Object Classes
                # +------+-------+-------+--------+
                # | Name | Shape | Color | Hotkey |
                # +------+-------+-------+--------+
                # +------+-------+-------+--------+
                # Tags
                # +------+------------+-----------------+--------+---------------+--------------------+
                # | Name | Value type | Possible values | Hotkey | Applicable to | Applicable classes |
                # +------+------------+-----------------+--------+---------------+--------------------+
                # +------+------------+-----------------+--------+---------------+--------------------+

                # Example 2: Complex ProjectMeta
                lemon = sly.ObjClass('lemon', sly.Rectangle)
                kiwi = sly.ObjClass('kiwi', sly.Polygon)
                tag_fruit = sly.TagMeta('fruit', sly.TagValueType.ANY_STRING)
                objects = sly.ObjClassCollection([lemon, kiwi])
                # or objects = [lemon, kiwi]
                tags = sly.TagMetaCollection([tag_fruit])
                # or tags = [tag_fruit]
                meta = sly.ProjectMeta(obj_classes=objects, tag_metas=tags, project_type=sly.ProjectType.IMAGES)
                print(meta)
                # Output:
                # +-------+-----------+----------------+--------+
                # |  Name |   Shape   |     Color      | Hotkey |
                # +-------+-----------+----------------+--------+
                # | lemon | Rectangle | [108, 15, 138] |        |
                # |  kiwi |  Polygon  | [15, 98, 138]  |        |
                # +-------+-----------+----------------+--------+
                # Tags
                # +-------+------------+-----------------+--------+---------------+--------------------+
                # |  Name | Value type | Possible values | Hotkey | Applicable to | Applicable classes |
                # +-------+------------+-----------------+--------+---------------+--------------------+
                # | fruit | any_string |       None      |        |      all      |         []         |
                # +-------+------------+-----------------+--------+---------------+--------------------+

                # Example 3: Add multi-view to the project
                lemon = sly.ObjClass('lemon', sly.Rectangle)
                kiwi = sly.ObjClass('kiwi', sly.Polygon)
                tag_fruit = sly.TagMeta('fruit', sly.TagValueType.ANY_STRING)

                settings = sly.ProjectSettings(
                    multiview_enabled=True,
                    multiview_tag_name=tag_fruit.name,
                    multiview_is_synced=False,
                )
                meta = sly.ProjectMeta(
                    obj_classes=[lemon, kiwi],
                    tag_metas=tag_fruit,
                    project_type=sly.ProjectType.IMAGES,
                    project_settings=settings
                )

                # Example 4: Custom color
                cat_class = sly.ObjClass("cat", sly.Rectangle, color=[0, 255, 0])
                scene_tag = sly.TagMeta("scene", sly.TagValueType.ANY_STRING)
                meta = sly.ProjectMeta(obj_classes=[cat_class], tag_metas=[scene_tag])
        """
        if obj_classes is None:
            self._obj_classes = ObjClassCollection()
        elif isinstance(obj_classes, list):
            self._obj_classes = ObjClassCollection(obj_classes)
        elif isinstance(obj_classes, ObjClassCollection):
            self._obj_classes = obj_classes
        else:
            raise TypeError(f"obj_classes argument has unknown type {type(obj_classes)}")

        if tag_metas is None:
            self._tag_metas = TagMetaCollection()
        elif isinstance(tag_metas, list):
            self._tag_metas = TagMetaCollection(tag_metas)
        elif isinstance(tag_metas, TagMetaCollection):
            self._tag_metas = tag_metas
        else:
            raise TypeError(f"tag_metas argument has unknown type {type(tag_metas)}")

        self._project_type = project_type

        if isinstance(project_settings, dict):
            self._project_settings = ProjectSettings.from_json(project_settings)
        else:
            self._project_settings = project_settings
            if project_settings is None:
                self._project_settings = ProjectSettings()

        self._project_settings.validate(self)

    @property
    def obj_classes(self) -> ObjClassCollection:
        """
        Collection of ObjClasses in ProjectMeta.

        :returns: ObjClassCollection object
        :rtype: :class:`~supervisely.annotation.obj_class_collection.ObjClassCollection`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                lemon = sly.ObjClass('lemon', sly.Rectangle)
                kiwi = sly.ObjClass('kiwi', sly.Polygon)
                objects = sly.ObjClassCollection([lemon, kiwi])
                # or objects = [lemon, kiwi]

                meta = sly.ProjectMeta(obj_classes=objects, project_type=sly.ProjectType.IMAGES)

                meta_classes = meta.obj_classes
                print(meta_classes.to_json())
                # Output: [
                #     {
                #         "title":"lemon",
                #         "shape":"rectangle",
                #         "color":"#6C0F8A",
                #         "geometry_config":{
                #
                #         },
                #         "hotkey":""
                #     },
                #     {
                #         "title":"kiwi",
                #         "shape":"polygon",
                #         "color":"#0F628A",
                #         "geometry_config":{
                #
                #         },
                #         "hotkey":""
                #     }
                # ]
        """
        return self._obj_classes

    @property
    def tag_metas(self) -> TagMetaCollection:
        """
        Collection of TagMetas in ProjectMeta.

        :returns: TagMetaCollection object
        :rtype: :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                tag_fruit = sly.TagMeta('fruit', sly.TagValueType.ANY_STRING)
                tags = sly.TagMetaCollection([tag_fruit])
                # or tags = [tag_fruit]

                meta = sly.ProjectMeta(tag_metas=tags)

                meta_tags = meta.tag_metas
                print(meta_tags.to_json())
                # Output: [
                #     {
                #         "name":"fruit",
                #         "value_type":"any_string",
                #         "color":"#818A0F",
                #         "hotkey":"",
                #         "applicable_type":"all",
                #         "classes":[]
                #     }
                # ]
        """
        return self._tag_metas

    @property
    def project_type(self) -> str:
        """
        Type of project. See possible value types in :class:`~supervisely.project.project_type.ProjectType`.

        :returns: Project type
        :rtype: str

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta = sly.ProjectMeta(project_type=sly.ProjectType.IMAGES)

                print(meta.project_type)
                # Output: 'images'
        """
        return self._project_type

    @property
    def project_settings(self) -> ProjectSettings:
        """
        Settings of the project. See possible values in :class:`~supervisely.project.project_settings.ProjectSettings`.

        :returns: Project settings
        :rtype: :class:`~supervisely.project.project_settings.ProjectSettings`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                s = sly.ProjectSettings(
                    multiview_enabled=True,
                    multiview_tag_name='multi_tag',
                    multiview_is_synced=False,
                )
                meta = sly.ProjectMeta(project_settings=s)

                print(meta.project_settings)
                # Output: <class 'supervisely.project.project_settings.ProjectSettings'>
        """
        return self._project_settings

    @property
    def labeling_interface(self) -> Optional[LabelingInterface]:
        """
        Get labeling interface settings of the project.

        :returns: Labeling interface settings
        :rtype: :class:`~supervisely.project.project_settings.LabelingInterface` or None

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                s = sly.ProjectSettings(
                    multiview_enabled=True,
                    multiview_tag_name='multi_tag',
                    multiview_is_synced=False,
                )
                meta = sly.ProjectMeta(project_settings=s)

                labeling_interface = meta.labeling_interface
                print(labeling_interface)
                # Output: None
        """
        return self._project_settings.labeling_interface if self._project_settings else None

    def to_json(self) -> Dict[str, Any]:
        """
        Convert the ProjectMeta to a json dict. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :returns: Json format as a dict
        :rtype: Dict[str, Any]

        :Usage Example:

            .. code-block:: python

                meta_json = meta.to_json()
                print(meta_json)
                # Output: {
                #     "classes": [
                #         {
                #             "title": "lemon",
                #             "shape": "rectangle",
                #             "color": "#720F8A",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         },
                #         {
                #             "title": "kiwi",
                #             "shape": "polygon",
                #             "color": "#8A0F6F",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         }
                #     ],
                #     "tags": [
                #         {
                #             "name": "fruit",
                #             "value_type": "any_string",
                #             "color": "#788A0F",
                #             "hotkey": "",
                #             "applicable_type": "all",
                #             "classes": []
                #         }
                #     ]
                # }
        """
        res = {
            ProjectMetaJsonFields.OBJ_CLASSES: self._obj_classes.to_json(),
            ProjectMetaJsonFields.TAGS: self._tag_metas.to_json(),
        }
        if self._project_type is not None:
            res[ProjectMetaJsonFields.PROJECT_TYPE] = self._project_type
        if self._project_settings is not None:
            res[ProjectMetaJsonFields.PROJECT_SETTINGS] = self._project_settings.to_json()
        return res

    @classmethod
    def from_json(cls, data: Dict) -> ProjectMeta:
        """
        Convert a json dict to ProjectMeta. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: ProjectMeta in json format as a dict.
        :type data: dict
        :returns: ProjectMeta object
        :rtype: :class:`~supervisely.project.project_meta.ProjectMeta`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta_json = {
                    "classes": [
                        {
                            "title": "lemon",
                            "shape": "rectangle",
                            "color": "#8A0F7B"
                        },
                        {
                            "title": "kiwi",
                            "shape": "polygon",
                            "color": "#8A0F50"
                        }
                    ],
                    "tags": [
                        {
                            "name": "fruit",
                            "value_type": "any_string",
                            "color": "#0F6F8A"
                        }
                    ]
                }
                meta = sly.ProjectMeta.from_json(meta_json)
        """
        tag_metas_json = data.get(ProjectMetaJsonFields.TAGS, [])
        img_tag_metas_json = data.get(ProjectMetaJsonFields.IMG_TAGS, [])
        obj_tag_metas_json = data.get(ProjectMetaJsonFields.OBJ_TAGS, [])
        project_type = data.get(ProjectMetaJsonFields.PROJECT_TYPE, None)

        if len(tag_metas_json) > 0:
            # New format - all project tags in a single collection.
            if any(len(x) > 0 for x in [img_tag_metas_json, obj_tag_metas_json]):
                raise ValueError(
                    "Project meta JSON contains both the {!r} section (current format merged tags for all of "
                    "the project) and {!r} or {!r} sections (legacy format with separate collections for images "
                    "and labeled objects). Either new format only or legacy format only are supported, but not a "
                    "mix.".format(
                        ProjectMetaJsonFields.TAGS,
                        ProjectMetaJsonFields.IMG_TAGS,
                        ProjectMetaJsonFields.OBJ_TAGS,
                    )
                )
            tag_metas = TagMetaCollection.from_json(tag_metas_json)
        else:
            img_tag_metas = TagMetaCollection.from_json(img_tag_metas_json)
            obj_tag_metas = TagMetaCollection.from_json(obj_tag_metas_json)
            tag_metas = _merge_img_obj_tag_metas(img_tag_metas, obj_tag_metas)

        obj_classes_json = data.get(ProjectMetaJsonFields.OBJ_CLASSES, None)
        if obj_classes_json is None:
            raise KeyError(
                f"Key '{ProjectMetaJsonFields.OBJ_CLASSES}' with the list of annotation classes "
                "not found in meta.json file. Check the annotation format documentation at: "
                "https://developer.supervisely.com/api-references/supervisely-annotation-json-format/project-classes-and-tags"
            )
        try:
            obj_classes = ObjClassCollection.from_json(obj_classes_json)
        except Exception as e:
            raise Exception(
                "Failed to deserialize classes from Project meta JSON. "
                "Check the annotation format documentation at: "
                "https://developer.supervisely.com/api-references/supervisely-annotation-json-format/project-classes-and-tags"
            ) from e

        project_settings_json = data.get(ProjectMetaJsonFields.PROJECT_SETTINGS, None)
        if project_settings_json is None:
            project_settings = None
        else:
            try:
                project_settings = ProjectSettings.from_json(project_settings_json)
            except Exception as e:
                raise Exception(
                    "Failed to deserialize settings from Project meta JSON. "
                    "Check the annotation format documentation at: "  # TODO
                    "https://developer.supervisely.com/api-references/supervisely-annotation-json-format/project-classes-and-tags"
                ) from e

        return cls(
            obj_classes=obj_classes,
            tag_metas=tag_metas,
            project_type=project_type,
            project_settings=project_settings,
        )

    def merge(self, other: ProjectMeta) -> ProjectMeta:
        """
        Merge all instances from given ProjectMeta into a single ProjectMeta object.

        :param other: Other project's meta.
        :type other: :class:`~supervisely.project.project_meta.ProjectMeta`
        :returns: New instance of project meta object
        :rtype: :class:`~supervisely.project.project_meta.ProjectMeta`
        :raises ValueError: Upon attempt to merge metas which contain the same obj class or tag meta

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta_1 = sly.ProjectMeta()
                class_cat = sly.ObjClass('cat', sly.Rectangle)
                tag_cat = sly.TagMeta('cat_tag', sly.TagValueType.ANY_STRING)
                meta_1 = meta_1.add_obj_class(class_cat)
                meta_1 = meta_1.add_tag_meta(tag_cat)

                meta_2 = sly.ProjectMeta()
                class_dog = sly.ObjClass('dog', sly.Rectangle)
                tag_dog = sly.TagMeta('dog_tag', sly.TagValueType.ANY_STRING)
                meta_2 = meta_2.add_obj_class(class_dog)
                meta_2 = meta_2.add_tag_meta(tag_dog)

                merge_meta = meta_1.merge(meta_2)
                merge_meta_json = merge_meta.to_json()
                print(json.dumps(merge_meta_json, indent=4))
                # Output: {
                #     "classes": [
                #         {
                #             "title": "dog",
                #             "shape": "rectangle",
                #             "color": "#0F8A62",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         },
                #         {
                #             "title": "cat",
                #             "shape": "rectangle",
                #             "color": "#340F8A",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         }
                #     ],
                #     "tags": [
                #         {
                #             "name": "dog_tag",
                #             "value_type": "any_string",
                #             "color": "#380F8A",
                #             "hotkey": "",
                #             "applicable_type": "all",
                #             "classes": []
                #         },
                #         {
                #             "name": "cat_tag",
                #             "value_type": "any_string",
                #             "color": "#8A0F82",
                #             "hotkey": "",
                #             "applicable_type": "all",
                #             "classes": []
                #         }
                #     ]
                # }
        """
        return self.clone(
            obj_classes=self._obj_classes.merge(other.obj_classes),
            tag_metas=self._tag_metas.merge(other._tag_metas),
        )

    def clone(
        self,
        obj_classes: Optional[Union[ObjClassCollection, List[ObjClass]]] = None,
        tag_metas: Optional[Union[TagMetaCollection, List[TagMeta]]] = None,
        project_type: Optional[str] = None,
        project_settings: Optional[Union[dict, ProjectSettings]] = None,
    ) -> ProjectMeta:
        """
        Clone makes a copy of ProjectMeta with new fields, if fields are given, otherwise it will use original ProjectMeta fields.

        :param obj_classes: ObjClassCollection or just list that stores :class:`~supervisely.annotation.obj_class.ObjClass` instances with unique names.
        :type obj_classes: :class:`~supervisely.annotation.obj_class_collection.ObjClassCollection` or List[:class:`~supervisely.annotation.obj_class.ObjClass`], optional
        :param tag_metas: TagMetaCollection that stores :class:`~supervisely.annotation.tag_meta.TagMeta` instances with unique names.
        :type tag_metas: :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection` or List[:class:`~supervisely.annotation.tag_meta.TagMeta`], optional
        :param project_type: Type of items in project: images, videos, volumes, point_clouds.
        :type project_type: str, optional
        :param project_settings: Additional project properties. For example, multi-view settings
        :type project_settings: dict or :class:`~supervisely.project.project_settings.ProjectSettings`, optional

        :returns: New instance of ProjectMeta object
        :rtype: :class:`~supervisely.project.project_meta.ProjectMeta`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta = sly.ProjectMeta()
                class_cat = sly.ObjClass('cat', sly.Rectangle)
                collection_cat = sly.ObjClassCollection([class_cat])
                # or collection_cat = [class_cat]
                tag_cat = sly.TagMeta('cat_tag', sly.TagValueType.ANY_STRING)
                collection_tag_cat = sly.TagMetaCollection([tag_cat])
                # or collection_tag_cat = [tag_cat]
                # Remember that ProjectMeta object is immutable, and we need to assign new instance of ProjectMeta to a new variable
                new_meta = meta.clone(obj_classes=collection_cat, tag_metas=collection_tag_cat)
                new_meta_json = new_meta.to_json()
                print(json.dumps(new_meta_json, indent=4))
                # Output: {
                #     "classes": [
                #         {
                #             "title": "cat",
                #             "shape": "rectangle",
                #             "color": "#190F8A",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         }
                #     ],
                #     "tags": [
                #         {
                #             "name": "cat_tag",
                #             "value_type": "any_string",
                #             "color": "#8A6D0F",
                #             "hotkey": "",
                #             "applicable_type": "all",
                #             "classes": []
                #         }
                #     ]
                # }
        """
        return ProjectMeta(
            obj_classes=take_with_default(obj_classes, self.obj_classes),
            tag_metas=take_with_default(tag_metas, self.tag_metas),
            project_type=take_with_default(project_type, self.project_type),
            project_settings=take_with_default(project_settings, self.project_settings),
        )

    def add_obj_class(self, new_obj_class: ObjClass) -> ProjectMeta:
        """
        Adds given ObjClass to ProjectMeta.

        :param new_obj_class: ObjClass object.
        :type new_obj_class: :class:`~supervisely.annotation.obj_class.ObjClass`
        :returns: New instance of ProjectMeta object
        :rtype: :class:`~supervisely.project.project_meta.ProjectMeta`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta = sly.ProjectMeta()
                class_cat = sly.ObjClass('cat', sly.Rectangle)
                # Remember that ProjectMeta object is immutable, and we need to assign new instance of ProjectMeta to a new variable
                meta = meta.add_obj_class(class_cat)
                meta_json = meta.to_json()
                print(json.dumps(meta_json, indent=4))
                # Output: {
                #     "classes": [
                #         {
                #             "title": "cat",
                #             "shape": "rectangle",
                #             "color": "#178A0F",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         }
                #     ],
                #     "tags": []
                # }
        """
        return self.add_obj_classes([new_obj_class])

    def add_obj_classes(
        self, new_obj_classes: Union[List[ObjClass], ObjClassCollection]
    ) -> ProjectMeta:
        """
        Adds given ObjClasses to ProjectMeta.

        :param new_obj_classes: List of ObjClass objects or ObjClassCollection.
        :type new_obj_classes: List[:class:`~supervisely.annotation.obj_class.ObjClass`] or :class:`~supervisely.annotation.obj_class_collection.ObjClassCollection`
        :returns: New instance of ProjectMeta object
        :rtype: :class:`~supervisely.project.project_meta.ProjectMeta`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta = sly.ProjectMeta()
                class_cat = sly.ObjClass('cat', sly.Rectangle)
                class_dog = sly.ObjClass('dog', sly.Bitmap)
                # Remember that ProjectMeta object is immutable, and we need to assign new instance of ProjectMeta to a new variable
                meta = meta.add_obj_classes([class_cat, class_dog])
                meta_json = meta.to_json()
                print(json.dumps(meta_json, indent=4))
                # Output: {
                #     "classes": [
                #         {
                #             "title": "cat",
                #             "shape": "rectangle",
                #             "color": "#8A0F3F",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         },
                #         {
                #             "title": "dog",
                #             "shape": "bitmap",
                #             "color": "#8A0F56",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         }
                #     ],
                #     "tags": []
                # }
        """
        return self.clone(obj_classes=self.obj_classes.add_items(new_obj_classes))

    def add_tag_meta(self, new_tag_meta: TagMeta) -> ProjectMeta:
        """
        Adds given TagMeta to ProjectMeta.

        :param new_tag_meta: TagMeta object.
        :type new_tag_meta: :class:`~supervisely.annotation.tag_meta.TagMeta`
        :returns: New instance of ProjectMeta object
        :rtype: :class:`~supervisely.project.project_meta.ProjectMeta`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta = sly.ProjectMeta()
                tag_cat = sly.TagMeta('cat_tag', sly.TagValueType.ANY_STRING)
                # Remember that ProjectMeta object is immutable, and we need to assign new instance of ProjectMeta to a new variable
                meta = meta.add_tag_meta(tag_cat)
                meta_json = meta.to_json()
                print(json.dumps(meta_json, indent=4))
                # Output: {
                #     "classes": [],
                #     "tags": [
                #         {
                #             "name": "cat_tag",
                #             "value_type": "any_string",
                #             "color": "#178A0F",
                #             "hotkey": "",
                #             "applicable_type": "all",
                #             "classes": []
                #         }
                #     ]
                # }
        """
        return self.add_tag_metas([new_tag_meta])

    def add_tag_metas(self, new_tag_metas: List[TagMeta]) -> ProjectMeta:
        """
        Adds given TagMetas to ProjectMeta.

        :param new_tag_metas: List of TagMeta objects or TagMetaCollection.
        :type new_tag_metas: List[:class:`~supervisely.annotation.tag_meta.TagMeta`] or :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection`
        :returns: New instance of ProjectMeta object
        :rtype: :class:`~supervisely.project.project_meta.ProjectMeta`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta = sly.ProjectMeta()
                tag_cat = sly.TagMeta('cat_tag', sly.TagValueType.ANY_STRING)
                tag_dog = sly.TagMeta('dog_tag', sly.TagValueType.ANY_STRING)
                # Remember that ProjectMeta object is immutable, and we need to assign new instance of ProjectMeta to a new variable
                meta = meta.add_tag_metas([tag_cat, tag_dog])
                meta_json = meta.to_json()
                print(json.dumps(meta_json, indent=4))
                # Output: {
                #     "classes": [],
                #     "tags": [
                #         {
                #             "name": "cat_tag",
                #             "value_type": "any_string",
                #             "color": "#0F248A",
                #             "hotkey": "",
                #             "applicable_type": "all",
                #             "classes": []
                #         },
                #         {
                #             "name": "dog_tag",
                #             "value_type": "any_string",
                #             "color": "#8A5C0F",
                #             "hotkey": "",
                #             "applicable_type": "all",
                #             "classes": []
                #         }
                #     ]
                # }
        """
        return self.clone(tag_metas=self.tag_metas.add_items(new_tag_metas))

    @staticmethod
    def _delete_items(collection, item_names):
        """
        :param collection: ObjClassCollection or TagMetaCollection instance
        :type collection: :class:`~supervisely.annotation.obj_class_collection.ObjClassCollection` or :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection`
        :param item_names: List of item names to delete
        :type item_names: List[str]
        :returns: List of items, which are in collection and not in given list of items to delete
        :rtype: List[:class:`~supervisely.annotation.obj_class.ObjClass`] or List[:class:`~supervisely.annotation.tag_meta.TagMeta`]
        """
        names_to_delete = set(item_names)
        res_items = []
        for item in collection:
            if item.key() not in names_to_delete:
                res_items.append(item)
        return res_items

    def delete_obj_class(self, obj_class_name: str) -> ProjectMeta:
        """
        Removes given ObjClass by name from ProjectMeta.

        :param obj_class_name: ObjClass name.
        :type obj_class_name: str
        :returns: New instance of ProjectMeta object
        :rtype: :class:`~supervisely.project.project_meta.ProjectMeta`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta = sly.ProjectMeta()
                class_cat = sly.ObjClass('cat', sly.Rectangle)
                # Remember that ProjectMeta object is immutable, and we need to assign new instance of ProjectMeta to a new variable
                meta = meta.add_obj_class(class_cat)
                meta_json = meta.to_json()
                print(json.dumps(meta_json, indent=4))
                # Output: {
                #     "classes": [
                #         {
                #             "title": "cat",
                #             "shape": "rectangle",
                #             "color": "#268A0F",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         }
                #     ],
                #     "tags": []
                # }

                # Remember that ProjectMeta object is immutable, and we need to assign new instance of ProjectMeta to a new variable
                meta = meta.delete_obj_class('cat')
                meta_json = meta.to_json()
                print(json.dumps(meta_json, indent=4))
                # Output: {
                #     "classes": [],
                #     "tags": []
                # }
        """
        return self.delete_obj_classes([obj_class_name])

    def delete_obj_classes(self, obj_class_names: List[str]) -> ProjectMeta:
        """
        Removes given ObjClasses by names from ProjectMeta.

        :param obj_class_names: List of ObjClasses names.
        :type obj_class_names: List[str]
        :returns: New instance of ProjectMeta object
        :rtype: :class:`~supervisely.project.project_meta.ProjectMeta`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta = sly.ProjectMeta()
                class_cat = sly.ObjClass('cat', sly.Rectangle)
                class_dog = sly.ObjClass('dog', sly.Bitmap)
                # Remember that ProjectMeta object is immutable, and we need to assign new instance of ProjectMeta to a new variable
                meta = meta.add_obj_classes([class_cat, class_dog])
                meta_json = meta.to_json()
                print(json.dumps(meta_json, indent=4))
                # Output: {
                #     "classes": [
                #         {
                #             "title": "cat",
                #             "shape": "rectangle",
                #             "color": "#8A0F18",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         },
                #         {
                #             "title": "dog",
                #             "shape": "bitmap",
                #             "color": "#0F8A7F",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         }
                #     ],
                #     "tags": []
                # }

                # Remember that ProjectMeta object is immutable, and we need to assign new instance of ProjectMeta to a new variable
                meta = meta.delete_obj_classes(['cat', 'dog'])
                meta_json = meta.to_json()
                print(json.dumps(meta_json, indent=4))
                # Output: {
                #     "classes": [],
                #     "tags": []
                # }
        """
        res_items = self._delete_items(self._obj_classes, obj_class_names)
        return self.clone(obj_classes=ObjClassCollection(res_items))

    def delete_tag_meta(self, tag_name: str) -> ProjectMeta:
        """
        Removes given TagMeta by name from ProjectMeta.

        :param tag_name: TagMeta name.
        :type tag_name: str
        :returns: New instance of ProjectMeta object
        :rtype: :class:`~supervisely.project.project_meta.ProjectMeta`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta = sly.ProjectMeta()
                tag_cat = sly.TagMeta('cat_tag', sly.TagValueType.ANY_STRING)
                # Remember that ProjectMeta object is immutable, and we need to assign new instance of ProjectMeta to a new variable
                meta = meta.add_tag_meta(tag_cat)
                meta_json = meta.to_json()
                print(json.dumps(meta_json, indent=4))
                # Output: {
                #     "classes": [],
                #     "tags": [
                #         {
                #             "name": "cat_tag",
                #             "value_type": "any_string",
                #             "color": "#8A540F",
                #             "hotkey": "",
                #             "applicable_type": "all",
                #             "classes": []
                #         }
                #     ]
                # }

                # Remember that ProjectMeta object is immutable, and we need to assign new instance of ProjectMeta to a new variable
                meta = meta.delete_tag_meta('cat_tag')
                meta_json = meta.to_json()
                print(json.dumps(meta_json, indent=4))
                # Output: {
                #     "classes": [],
                #     "tags": []
                # }
        """
        return self.delete_tag_metas([tag_name])

    def delete_tag_metas(self, tag_names: List[str]) -> ProjectMeta:
        """
        Removes given TagMetas by names from ProjectMeta.

        :param tag_names: List of TagMetas names.
        :type tag_names: List[:class:`~supervisely.annotation.tag_meta.TagMeta`]
        :returns: New instance of ProjectMeta object
        :rtype: :class:`~supervisely.project.project_meta.ProjectMeta`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta = sly.ProjectMeta()
                tag_cat = sly.TagMeta('cat_tag', sly.TagValueType.ANY_STRING)
                tag_dog = sly.TagMeta('dog_tag', sly.TagValueType.ANY_STRING)

                # Remember that ProjectMeta object is immutable, and we need to assign new instance of ProjectMeta to a new variable
                meta = meta.add_tag_metas([tag_cat, tag_dog])
                meta_json = meta.to_json()
                print(json.dumps(meta_json, indent=4))
                # Output: {
                #     "classes": [],
                #     "tags": [
                #         {
                #             "name": "cat_tag",
                #             "value_type": "any_string",
                #             "color": "#0F298A",
                #             "hotkey": "",
                #             "applicable_type": "all",
                #             "classes": []
                #         },
                #         {
                #             "name": "dog_tag",
                #             "value_type": "any_string",
                #             "color": "#8A410F",
                #             "hotkey": "",
                #             "applicable_type": "all",
                #             "classes": []
                #         }
                #     ]
                # }

                # Remember that ProjectMeta object is immutable, and we need to assign new instance of ProjectMeta to a new variable
                meta = meta.delete_tag_metas(['cat_tag', 'dog_tag'])
                meta_json = meta.to_json()
                print(json.dumps(meta_json, indent=4))
                # Output: {
                #     "classes": [],
                #     "tags": []
                # }
        """
        res_items = self._delete_items(self._tag_metas, tag_names)
        return self.clone(tag_metas=TagMetaCollection(res_items))

    def get_obj_class(self, obj_class_name: str) -> ObjClass:
        """
        Get given ObjClass by name from ProjectMeta.

        :param obj_class_name: ObjClass name.
        :type obj_class_name: str
        :returns: :class:`~supervisely.annotation.obj_class.ObjClass` object
        :rtype: :class:`~supervisely.annotation.obj_class.ObjClass`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta = sly.ProjectMeta()
                class_cat = sly.ObjClass('cat', sly.Rectangle)
                class_dog = sly.ObjClass('dog', sly.Bitmap)
                # Remember that ProjectMeta object is immutable, and we need to assign new instance of ProjectMeta to a new variable
                meta = meta.add_obj_classes([class_cat, class_dog])
                meta_json = meta.to_json()
                print(json.dumps(meta_json, indent=4))
                # Output: {
                #     "classes": [
                #         {
                #             "title": "cat",
                #             "shape": "rectangle",
                #             "color": "#8A140F",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         },
                #         {
                #             "title": "dog",
                #             "shape": "bitmap",
                #             "color": "#0F8A35",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         }
                #     ],
                #     "tags": []
                # }

                class_cat = meta.get_obj_class('cat')
                print(class_cat)
                # Output:
                # Name:  cat       Shape: Rectangle    Color: [138, 20, 15]  Geom. settings: {}              Hotkey

                class_elephant = meta.get_obj_class('elephant')
                print(class_elephant)
                # Output:
                # None
        """
        return self._obj_classes.get(obj_class_name)

    def get_obj_class_by_id(self, obj_class_id: int) -> Optional[ObjClass]:
        """
        Get given ObjClass by name from ProjectMeta.

        :param obj_class_id: ObjClass id.
        :type obj_class_id: int
        :returns: :class:`~supervisely.annotation.obj_class.ObjClass` object or None
        :rtype: :class:`~supervisely.annotation.obj_class.ObjClass`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                project_id = 555
                meta = sly.ProjectMeta.from_json(api.project.get_meta(project_id))
                obj_class_id = 123
        """
        for obj_class in self.obj_classes:
            if obj_class.sly_id == obj_class_id:
                return obj_class

    def get_tag_meta(self, tag_name: str) -> Optional[TagMeta]:
        """
        Get given TagMeta by name from ProjectMeta.

        :param tag_name: TagMeta name.
        :type tag_name: str
        :returns: :class:`~supervisely.annotation.tag_meta.TagMeta` object or None.
        :rtype: :class:`~supervisely.annotation.tag_meta.TagMeta`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta = sly.ProjectMeta()
                tag_cat = sly.TagMeta('cat_tag', sly.TagValueType.ANY_STRING)
                tag_dog = sly.TagMeta('dog_tag', sly.TagValueType.ANY_STRING)
                # Remember that ProjectMeta object is immutable, and we need to assign new instance of ProjectMeta to a new variable
                meta = meta.add_tag_metas([tag_cat, tag_dog])
                meta_json = meta.to_json()
                print(json.dumps(meta_json, indent=4))
                # Output: {
                #     "classes": [],
                #     "tags": [
                #         {
                #             "name": "cat_tag",
                #             "value_type": "any_string",
                #             "color": "#590F8A",
                #             "hotkey": "",
                #             "applicable_type": "all",
                #             "classes": []
                #         },
                #         {
                #             "name": "dog_tag",
                #             "value_type": "any_string",
                #             "color": "#0F8A88",
                #             "hotkey": "",
                #             "applicable_type": "all",
                #             "classes": []
                #         }
                #     ]
                # }

                tag_cat = meta.get_tag_meta('cat_tag')
                print(tag_cat)
                # Output:
                # Name:  cat_tag                  Value type:any_string    Possible values:None       Hotkey                  Applicable toall        Applicable classes[]

                tag_elephant = meta.get_tag_meta('elephant_tag')
                print(tag_elephant)
                # Output:
                # None
        """
        return self._tag_metas.get(tag_name)

    def get_tag_meta_by_id(self, tag_id: int) -> Optional[TagMeta]:
        """Return TagMeta with given id.

        :param tag_id: TagMeta id to search for.
        :type tag_id: int
        :returns: TagMeta with given id.
        :rtype: :class:`~supervisely.annotation.tag_meta.TagMeta` or None
        """
        return self._tag_metas.get_by_id(tag_id)

    def get_tag_name_by_id(self, tag_id: int) -> Optional[str]:
        """Return tag name with given id.

        :param tag_id: TagMeta id to search for.
        :type tag_id: int
        :returns: tag name with given id.
        :rtype: tag name or None
        """
        return self._tag_metas.get_tag_name_by_id(tag_id)

    @staticmethod
    def merge_list(metas: List[ProjectMeta]) -> ProjectMeta:
        """
        Merge ProjectMetas from given list of ProjectMetas into single ProjectMeta object.

        :param metas: List of :class:`~supervisely.project.project_meta.ProjectMeta` objects.
        :type metas: List[:class:`~supervisely.project.project_meta.ProjectMeta`]
        :returns: New instance of :class:`~supervisely.project.project_meta.ProjectMeta` object
        :rtype: :class:`~supervisely.project.project_meta.ProjectMeta`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta = sly.ProjectMeta()

                meta_1 = sly.ProjectMeta()
                class_cat = sly.ObjClass('cat', sly.Rectangle)
                meta_1 = meta_1.add_obj_class(class_cat)

                meta_2 = sly.ProjectMeta()
                tag_dog = sly.TagMeta('dog_tag', sly.TagValueType.ANY_STRING)
                meta_2 = meta_2.add_tag_meta(tag_dog)

                # Remember that ProjectMeta object is immutable, and we need to assign new instance of ProjectMeta to a new variable
                meta = meta.merge_list([meta_1, meta_2])
                meta_json = meta.to_json()
                print(json.dumps(meta_json, indent=4))
                # Output: {
                #     "classes": [
                #         {
                #             "title": "cat",
                #             "shape": "rectangle",
                #             "color": "#0F8A45",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         }
                #     ],
                #     "tags": [
                #         {
                #             "name": "dog_tag",
                #             "value_type": "any_string",
                #             "color": "#320F8A",
                #             "hotkey": "",
                #             "applicable_type": "all",
                #             "classes": []
                #         }
                #     ]
                # }
        """
        res_meta = ProjectMeta()
        for meta in metas:
            res_meta = res_meta.merge(meta)
        return res_meta

    def __str__(self):
        result = "ProjectMeta:\n"
        result += "Object Classes\n{}\n".format(str(self._obj_classes))
        result += "Tags\n{}\n".format(str(self._tag_metas))
        return result

    def __eq__(self, other: ProjectMeta):
        if self.obj_classes == other.obj_classes and self.tag_metas == other.tag_metas:
            return True
        return False

    def __ne__(self, other: ProjectMeta):
        return not self == other

    def to_segmentation_task(
        self, keep_geometries: Optional[List] = [Polygon, Bitmap], target_classes=None
    ) -> Tuple[ProjectMeta, Dict[ObjClass, ObjClass]]:
        """
        Convert project meta classes geometries with keep_geometries types to Bitmaps and create new ProjectMeta.

        :param keep_geometries: List of geometries that can be converted.
        :type keep_geometries: List, optional
        :returns: New project meta and dict correspondences of old classes to new
        :rtype: Tuple[:class:`~supervisely.project.project_meta.ProjectMeta`, Dict[:class:`~supervisely.annotation.obj_class.ObjClass`, :class:`~supervisely.annotation.obj_class.ObjClass`]]

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta = sly.ProjectMeta()
                class_cat = sly.ObjClass('cat', sly.Polygon)
                class_dog = sly.ObjClass('dog', sly.Bitmap)
                meta = meta.add_obj_classes([class_cat, class_dog])
                meta_json = meta.to_json()
                print(json.dumps(meta_json, indent=4))
                # Output: {
                #     "classes": [
                #         {
                #             "title": "cat",
                #             "shape": "polygon",
                #             "color": "#208A0F",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         },
                #         {
                #             "title": "dog",
                #             "shape": "bitmap",
                #             "color": "#8A570F",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         }
                #     ],
                #     "tags": []
                # }

                res_meta, mapping = meta.to_segmentation_task()
                res_meta_json = res_meta.to_json()
                print(json.dumps(res_meta_json, indent=4))
                # Output: {
                #     "classes": [
                #         {
                #             "title": "cat",
                #             "shape": "bitmap",
                #             "color": "#208A0F",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         },
                #         {
                #             "title": "dog",
                #             "shape": "bitmap",
                #             "color": "#8A570F",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         }
                #     ],
                #     "tags": []
                # }
        """
        mapping = {}
        res_classes = []
        for obj_class in self.obj_classes:
            obj_class: ObjClass

            if target_classes is None or obj_class.name in target_classes:
                if obj_class.geometry_type in keep_geometries:
                    if obj_class.geometry_type == Bitmap:
                        mapping[obj_class] = obj_class
                        res_classes.append(obj_class)
                    else:
                        new_class = obj_class.clone(geometry_type=Bitmap)
                        mapping[obj_class] = new_class
                        res_classes.append(new_class)
                else:
                    mapping[obj_class] = None
            else:
                mapping[obj_class] = None

        res_meta = self.clone(obj_classes=ObjClassCollection(res_classes))
        return res_meta, mapping

    def to_detection_task(
        self, convert_classes: Optional[bool] = False
    ) -> Tuple[ProjectMeta, Dict[ObjClass, ObjClass]]:
        """
        Convert project meta classes geometries to Rectangles or skip them and create new ProjectMeta.

        :param convert_classes: Convert classes with no Rectangle type to Rectangle or skip them.
        :type convert_classes: bool, optional
        :returns: New project meta and dict correspondences of old classes to new
        :rtype: :class:`Tuple[ProjectMeta, Dict[ObjClass, ObjClass]]`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta = sly.ProjectMeta()
                class_cat = sly.ObjClass('cat', sly.Polygon)
                class_dog = sly.ObjClass('dog', sly.Bitmap)
                meta = meta.add_obj_classes([class_cat, class_dog])
                meta_json = meta.to_json()
                print(json.dumps(meta_json, indent=4))
                # Output: {
                #     "classes": [
                #         {
                #             "title": "cat",
                #             "shape": "polygon",
                #             "color": "#208A0F",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         },
                #         {
                #             "title": "dog",
                #             "shape": "bitmap",
                #             "color": "#8A570F",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         }
                #     ],
                #     "tags": []
                # }

                res_meta, mapping = meta.to_detection_task(convert_classes=True)
                res_meta_json = res_meta.to_json()
                print(json.dumps(res_meta_json, indent=4))
                # Output: {
                #     "classes": [
                #         {
                #             "title": "cat",
                #             "shape": "rectangle",
                #             "color": "#3A0F8A",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         },
                #         {
                #             "title": "dog",
                #             "shape": "rectangle",
                #             "color": "#8A310F",
                #             "geometry_config": {},
                #             "hotkey": ""
                #         }
                #     ],
                #     "tags": []
                # }
        """
        mapping = {}
        res_classes = []
        for obj_class in self.obj_classes:
            obj_class: ObjClass
            if obj_class.geometry_type == Rectangle:
                mapping[obj_class] = obj_class
                res_classes.append(obj_class)
            else:
                if convert_classes is True:
                    new_class = obj_class.clone(geometry_type=Rectangle)
                    mapping[obj_class] = new_class
                    res_classes.append(new_class)
                else:
                    # ignore class
                    mapping[obj_class] = None
        res_meta = self.clone(obj_classes=ObjClassCollection(res_classes))
        return res_meta, mapping



from __future__ import annotations

from typing import Any, Dict, List, Optional

import supervisely.imaging.color as sly_color
from supervisely.app.widgets import Widget


class Field(Widget):
    """Form field with title, description, and optional icon; wraps another widget."""

    class Icon:
        """Material Design Icon or image for Field header."""

        def __init__(
            self,
            zmdi_class: Optional[str] = None,
            color_rgb: Optional[List[int, int, int]] = None,
            bg_color_rgb: Optional[List[int, int, int]] = None,
            image_url: Optional[str] = None,
        ) -> Field.Icon:
            """
            :param zmdi_class: Material Design Icon class name.
            :type zmdi_class: Optional[str]
            :param color_rgb: Icon color [R, G, B].
            :type color_rgb: Optional[List[int, int, int]]
            :param bg_color_rgb: Background color [R, G, B].
            :type bg_color_rgb: Optional[List[int, int, int]]
            :param image_url: Image URL (alternative to zmdi_class).
            :type image_url: Optional[str]
            :raises ValueError: If both or neither zmdi_class/image_url; invalid colors.
            """
            if zmdi_class is None and image_url is None:
                raise ValueError(
                    "One of the arguments has to be defined: zmdi_class or image_url"
                )
            if zmdi_class is not None and image_url is not None:
                raise ValueError(
                    "Only one of the arguments has to be defined: zmdi_class or image_url"
                )

            if image_url is None and color_rgb is None:
                color_rgb = [255, 255, 255]

            if image_url is None and bg_color_rgb is None:
                bg_color_rgb = [0, 154, 255]

            self._zmdi_class = zmdi_class
            self._color = color_rgb
            self._bg_color = bg_color_rgb
            self._image_url = image_url
            if self._color is not None:
                sly_color._validate_color(self._color)
            if self._bg_color is not None:
                sly_color._validate_color(self._bg_color)

        def to_json(self) -> Dict[str, Any]:
            """Returns JSON representation of the icon.

            Dictionary contains the following fields:
                If icon is Material Design Icon:
                    - className: Material Design Icon class name
                    - color: RGB color of the icon
                    - bgColor: RGB color of the icon background
                If icon is image:
                    - imageUrl: URL of the icon image

            :returns: JSON representation of the icon
            :rtype: Dict[str, Any]
            """
            res = {}
            if self._zmdi_class is not None:
                res["className"] = self._zmdi_class
                res["color"] = sly_color.rgb2hex(self._color)
                res["bgColor"] = sly_color.rgb2hex(self._bg_color)
            if self._image_url is not None:
                res["imageUrl"] = self._image_url
                res["bgColor"] = sly_color.rgb2hex(self._bg_color)
            return res

    def __init__(
        self,
        content: Widget,
        title: str,
        description: Optional[str] = None,
        title_url: Optional[str] = None,
        description_url: Optional[str] = None,
        icon: Optional[Field.Icon] = None,
        widget_id: Optional[str] = None,
    ):
        """
        :param content: Child widget.
        :type content: :class:`~supervisely.app.widgets.widget.Widget`
        :param title: Field title.
        :type title: str
        :param description: Field description.
        :type description: Optional[str]
        :param title_url: Link URL for title.
        :type title_url: Optional[str]
        :param description_url: Link URL for description.
        :type description_url: Optional[str]
        :param icon: Field.Icon for header.
        :type icon: Optional[Field.Icon]
        :param widget_id: Widget identifier.
        :type widget_id: Optional[str]

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import Field, Text
                field = Field(Text("Hello"), "Title", "Description", icon=Field.Icon(zmdi_class="zmdi zmdi-bike"))
        """
        self._title = title
        self._description = description
        self._title_url = title_url
        self._description_url = description_url
        self._icon = icon
        self._content = content
        if self._title_url is not None and self._title is None:
            raise ValueError(
                "Title can not be specified only as url without text value"
            )
        if self._description_url is not None and self._description is None:
            raise ValueError(
                "Description can not be specified only as url without text value"
            )

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict[str, Any]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.
        Dictionary contains the following fields:
            - title: Title of the field
            - description: Description of the field
            - title_url: URL of the title
            - description_url: URL of the description
            - icon: Icon for the field
            If icon is Material Design Icon:
            - icon with the following fields:
                - className: Material Design Icon class name
                - color: RGB color of the icon
                - bgColor: RGB color of the icon background

        :returns: Dictionary with widget data
        :rtype: Dict[str, Any]
        """
        res = {
            "title": self._title,
            "description": self._description,
            "title_url": self._title_url,
            "description_url": self._description_url,
            "icon": None,
        }
        if self._icon is not None:
            res["icon"] = self._icon.to_json()
        return res

    def get_json_state(self) -> None:
        """Field widget does not have state, the method returns None."""
        return None



"""This module contains the implementation of the TreeSelect widget in Supervisely."""

# Source documentation: https://vue-treeselect.js.org/

from __future__ import annotations

from typing import Callable, Dict, List, Optional, Union

from supervisely.app.content import DataJson, StateJson
from supervisely.app.widgets import Widget


class TreeSelect(Widget):
    """Dropdown for selecting items from a tree-like structure.
    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/selection/treeselect>`_.
    """

    class Routes:
        """Endpoints for the widget."""

        VALUE_CHANGED = "value_changed"

    class Item:
        """Class representing an item in the tree."""

        def __init__(
            self,
            id: str,
            label: Optional[str] = None,
            children: List[TreeSelect.Item] = None,
        ):
            """
            :param id: The unique identifier of the item.
            :type id: str
            :param label: The label of the item. If None, id is used.
            :type label: str, optional
            :param children: The children of the item.
            :type children: List[:class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`], optional

            :Usage Example:

                .. code-block:: python

                    TreeSelect.Item(id="1", label="First item", children=[
                        TreeSelect.Item(id="1.1", label="First child"),
                        TreeSelect.Item(id="1.2", label="Second child"),
                    ])
            """
            self.id = id
            self.label = label or id
            self.children = children or []

        def to_json(self) -> Dict[str, Union[str, List[Dict]]]:
            data = {
                "id": self.id,
                "label": self.label,
            }
            if self.children:
                data["children"] = [child.to_json() for child in self.children]

            return data

        @classmethod
        def from_json(cls, data: Dict[str, Union[str, List[Dict]]]) -> TreeSelect.Item:
            return cls(
                id=data["id"],
                label=data.get("label"),
                children=[TreeSelect.Item.from_json(child) for child in data.get("children", [])],
            )

        def __repr__(self):
            return f"Item(id={self.id}, label={self.label}, children={self.children})"

    def __init__(
        self,
        items: Optional[List[TreeSelect.Item]] = None,
        multiple_select: bool = False,
        flat: bool = False,
        always_open: bool = False,
        width: Optional[int] = None,
        append_to_body: bool = True,
        widget_id: Optional[str] = None,
        placeholder: Optional[str] = None,
        show_tooltip: bool = True,
    ):
        """Initialize the TreeSelect widget.

        :param items: List of items to display in the tree.
        :type items: List[:class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`], optional
        :param multiple_select: Whether multiple items can be selected at once.
        :type multiple_select: bool, optional
        :param flat: If True, selecting a parent will NOT select all its children.
        :type flat: bool, optional
        :param always_open: If True, the widget will be expanded by default.
        :type always_open: bool, optional
        :param width: The width of the widget.
        :type width: int, optional
        :param append_to_body: If False, popover is inside input container (may be hidden in Card).
        :type append_to_body: bool, optional
        :param widget_id: The unique identifier of the widget.
        :type widget_id: str, optional
        :param placeholder: The placeholder text.
        :type placeholder: str, optional
        :param show_tooltip: Whether to show tooltip.
        :type show_tooltip: bool

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import TreeSelect

                items = [
                    TreeSelect.Item(id="1", label="First item", children=[
                        TreeSelect.Item(id="1.1", label="First child"),
                        TreeSelect.Item(id="1.2", label="Second child"),
                    ]),
                    TreeSelect.Item(id="2", label="Second item"),
                ]
                tree_select = TreeSelect(multiple_select=True, flat=True, always_open=True)
                tree_select.set_items(items)
        """
        self._items = items or []
        self._multiple = multiple_select
        self._flat = flat
        self._always_open = always_open
        self._value_format = "object"  # On the frontend side can be "object" or "id".
        self._value = [] if multiple_select else None
        self._width = width
        self._append_to_body = append_to_body
        self._placeholder = placeholder
        self._show_tooltip = show_tooltip

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict[str, List[Dict]]:
        """Return the JSON representation of the widget data.

        :returns: The JSON representation of the widget data.
        :rtype: Dict[str, List[Dict]]
        """
        return {
            "items": [item.to_json() for item in self._items] if self._items else [],
            "width": self._width,
        }

    def get_json_state(self) -> Dict[str, Union[Dict, List[Dict]]]:
        """Return the JSON representation of the widget state.

        :returns: The JSON representation of the widget state.
        :rtype: Dict[str, Union[Dict, List[Dict]]]
        """
        return {
            "value": self.value,
            "options": {
                "multiple": self._multiple,
                "flat": self._flat,
                "alwaysOpen": self._always_open,
                "valueFormat": self._value_format,
                "appendToBody": self._append_to_body,
                "placeholder": self._placeholder,
                "showTooltip": self._show_tooltip,
            },
        }

    @property
    def value(self) -> Union[List[TreeSelect.Item], TreeSelect.Item]:
        """Return the selected item(s).
        Do not use this property directly, use get_selected() instead.

        :returns: The selected item(s).
        :rtype: Union[List[:class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`], :class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`]
        """
        return self._value

    @value.setter
    def value(self, value: Union[List[TreeSelect.Item], TreeSelect.Item]):
        """Set the selected item(s).
        Do not use this property directly, use set_selected() instead.

        :param value: The selected item(s).
        :type value: Union[List[:class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`], :class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`]
        """
        self._value = value

    def _get_value(self) -> Union[List[TreeSelect.Item], TreeSelect.Item]:
        """Get the selected item as instances of the Item class.

        :returns: The selected item(s).
        :rtype: Union[List[:class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`], :class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`]
        """
        res = StateJson()[self.widget_id]["value"]
        if res is None:
            return None
        if isinstance(res, list):
            return [TreeSelect.Item.from_json(item) for item in res]
        return TreeSelect.Item.from_json(res)

    def _set_value(self, value: Optional[Union[List[TreeSelect.Item], TreeSelect.Item]]):
        """Set the selected item(s) as instances of the Item class.

        :param value: The selected item(s).
        :type value: Union[List[:class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`], :class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`]
        """
        self.value = value
        if isinstance(value, list):
            json_value = [item.to_json() for item in value]
        elif value is not None:
            json_value = value.to_json()
        else:
            json_value = None if not self._multiple else []
        StateJson()[self.widget_id]["value"] = json_value
        StateJson().send_changes()

    def clear_selected(self) -> None:
        """Clear the selected item(s)."""
        self._set_value([] if self._multiple else None)

    def get_selected(self) -> Union[List[TreeSelect.Item], TreeSelect.Item]:
        """Get the selected item(s).

        :returns: The selected item(s).
        :rtype: Union[List[:class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`], :class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`]
        """
        return self._get_value()

    def set_selected(self, value: Union[List[TreeSelect.Item], TreeSelect.Item]):
        """Set the selected item(s).

        :param value: The selected item(s).
        :type value: Union[List[:class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`], :class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`]
        :raises ValueError: If the widget is set to single selection mode and a list of items
            is provided.
        """
        if self._multiple and not isinstance(value, list):
            value = [value]
        if not self._multiple and isinstance(value, list):
            raise ValueError(
                "The widget is set to single selection mode, but a list of items was provided."
                "Either set the widget to multiple selection mode or provide a single item."
            )

        self._set_value(value)

    def get_all_items(self) -> List[TreeSelect.Item]:
        """Get all items in the tree.

        :returns: All items in the tree.
        :rtype: List[:class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`]
        """

        def _get_all_items(items: List[TreeSelect.Item]) -> List[TreeSelect.Item]:
            res = []
            if not items:
                return res
            for item in items:
                res.append(item)
                res.extend(_get_all_items(item.children))
            return res

        return _get_all_items(self._items)

    def select_all(self) -> None:
        """Select all items, including children."""
        if not self._multiple:
            raise ValueError(
                "The widget is set to single selection mode, but tried to select all items."
            )
        self._set_value(self.get_all_items())

    def is_all_selected(self) -> bool:
        """Check if all items are selected.

        :returns: True if all items are selected, False otherwise.
        :rtype: bool
        """
        if not self._multiple:
            raise ValueError(
                "The widget is set to single selection mode, but tried to check if all items are selected."
            )

        all_item_ids = {item.id for item in self.get_all_items()}
        selected_ids = {item.id for item in self.get_selected()}

        return all_item_ids.issubset(selected_ids)

    def get_item_by_id(self, item_id: str) -> Optional[TreeSelect.Item]:
        """Get the item by its ID.

        :param item_id: The ID of the item.
        :type item_id: str
        :returns: The item with the specified ID.
        :rtype: Optional[:class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`]
        """

        def _get_item_by_id(
            items: List[TreeSelect.Item], item_id: str
        ) -> Optional[TreeSelect.Item]:
            for item in items:
                if item.id == item_id:
                    return item
                res = _get_item_by_id(item.children, item_id)
                if res:
                    return res
            return None

        return _get_item_by_id(self._items, item_id)

    def set_selected_by_id(self, value: Union[List[str], str]) -> None:
        if not self._multiple and isinstance(value, list):
            raise ValueError(
                "The widget is set to single selection mode, but a list of items was provided."
                "Either set the widget to multiple selection mode or provide a single item."
            )
        if not isinstance(value, list):
            value = [value]

        items = []
        for item_id in value:
            item = self.get_item_by_id(item_id)
            if item:
                items.append(item)

        if items:
            self.set_selected(items if self._multiple else items[0])

    def _update_items(
        self, items: Union[List[TreeSelect.Item], TreeSelect.Item], overwrite: bool
    ) -> None:
        """Update the items in the widget.

        :param items: The items to update.
        :type items: Union[List[:class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`], :class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`]
        :param overwrite: Whether to overwrite the existing items.
        :type overwrite: bool
        """
        if overwrite:
            self._items = items
            self.clear_selected()
        else:
            self._items.extend(items)
        DataJson()[self.widget_id] = self.get_json_data()
        DataJson().send_changes()

    def set_items(self, items: List[TreeSelect.Item]) -> None:
        """Set the items in the widget.

        :param items: The items to set.
        :type items: List[:class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`]
        """
        self._update_items(items, overwrite=True)

    def add_items(self, items: List[TreeSelect.Item]) -> None:
        """Add the items to the widget.

        :param items: The items to add.
        :type items: List[:class:`~supervisely.app.widgets.tree_select.tree_select.TreeSelect.Item`]
        """
        self._update_items(items, overwrite=False)

    def clear_items(self) -> None:
        """Clear the items in the widget."""
        items = [] if self._multiple else None
        self._update_items(items, overwrite=True)

    def value_changed(self, func: Callable) -> Callable:
        """Decorator for the function to be called when the value is changed.

        :param func: The function to be called when the value is changed.
        :type func: Callable
        :returns: The decorated function.
        :rtype: Callable
        """
        route_path = self.get_route_path(TreeSelect.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():
            res = self._get_value()

            func(res)

        return _click
